# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.

"""
Config parser module
"""

import configparser
import os
import sys
import types

from ctypes.util import find_library

import clang.cindex as cli
from cppbind.common import PROJECT_CONFIG_DIR
from .error import Error

PROJECT_CONFIG = os.path.join(PROJECT_CONFIG_DIR, "cppbind_config.cfg")

DEFAULT_DIRS = ['', './', PROJECT_CONFIG_DIR]

# find clang for windows and mac
clang_lib = find_library('libclang')

# find clang for linux
if clang_lib is None:
    for version in range(13, 5, -1):
        clang_lib = find_library(f'clang-{version}')
        if clang_lib:
            break

if clang_lib is None:
    print("clang dev is not installed. Please read README.md")
    sys.exit(1)

# setting clang library file.
# in feature we might  consider to have path optionally defined in config
cli.Config.set_library_file(clang_lib)


def read_config(config_file=None):
    """
    creates and loads config file.
    Args:
        config_file: default config file
    """
    config_file = config_file or PROJECT_CONFIG

    config = configparser.ConfigParser(
        converters={
            'list': lambda x: [i.strip() for i in x.split(',')]
        }
    )

    try:
        with open(config_file) as file:
            config.read_file(file)
    except OSError as err:
        Error.critical(f"Cannot read config file {config_file}: {err}")

    return config


class CppBindConfig:
    """
    Loads cppbind config file into structure
    """

    def __init__(self, file_names=None):
        """
        initializes config

        loads default config as well as config provided by user
        Args:
            file_names: list of file names that is going to be loaded
        """

        cnfg = read_config()

        if file_names is not None:
            cnfg.read(file_names)

        self.application = types.SimpleNamespace(**dict(cnfg.items('APPLICATION')))

        self.default_config_dirs = DEFAULT_DIRS

        self.languages = cnfg.getlist("APPLICATION", "all_languages")
        self.platforms = cnfg.getlist("APPLICATION", "all_platforms")

        self.api_start_kw = cnfg.get("API", "parser_start")

        self.logging = types.SimpleNamespace(**dict(cnfg.items("LOG")))

    def __repr__(self):
        return f"CppBind_Config({repr(self.__dict__)})"


config = CppBindConfig(["~/cppbind_config.cfg", "cppbind_config.cfg"])



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.


from cppbind.common.error import Error


def allowed_after_build(method):
    """Decorator to check that the method/property is not called before IR is built"""

    def wrapper(self, *args, **kwargs):
        if not self.root._is_built():
            Error.internal(f"IR is not completely built. Access to '{method.__name__}' property is forbidden")

        return method(self, *args, **kwargs)

    return wrapper


def available_on(*allowed_cursor_kinds):
    """Decorator to check that the method is called only for cursors with an allowed kind"""

    def decorator(method):

        def wrapper(self, *args, **kwargs):
            # for differentiating between Node and Context methods/properties
            kind = self.kind if hasattr(self, 'kind') else self.node.kind

            if kind not in allowed_cursor_kinds:
                raise AttributeError(f"{self.__class__.__name__}.{method.__name__} is "
                                     f"invalid for {kind} node kind.")

            return method(self, *args, **kwargs)

        return wrapper

    return decorator



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.

"""
"""

from abc import abstractmethod, ABC
from collections import OrderedDict
from enum import Enum

from cached_property import cached_property
from sortedcontainers import SortedSet

import clang.cindex as cli
import cppbind.utils.clang as cutil
from cppbind.ir.cxx_element import CXXElement
from cppbind.ir.cxx_type import CXXType
from . import ElementKind
from .utils import allowed_after_build, available_on


class NodeType(Enum):
    CLANG_NODE = 1
    DIRECTORY_NODE = 2
    ROOT_NODE = 3
    FILE_NODE = 4


ROOT_KIND_NAME = "root"
DIR_KIND_NAME = "dir"
FILE_KIND_NAME = "file"


class Node(ABC):
    API_NONE = 'none'

    def __init__(self, api=None, args=None, root=None, parent=None, children=None, pure_comment=None):
        self.api = api
        self.pure_comment = pure_comment
        self.args = args or OrderedDict()
        self.root = root
        self.parent = parent
        self._children = children or []

    def __eq__(self, other):
        return self.signature == other.signature

    def __lt__(self, other):
        return self.signature < other.signature

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.signature)

    def __repr__(self):
        return f"Node(api={self.api}, args={self.args} children={self.children})"

    @abstractmethod
    def __iter__(self):
        # each node is an iterator and defines its iteration order
        pass

    @abstractmethod
    def __next__(self):
        # each node is an iterator and defines its iteration order
        pass

    @property
    @abstractmethod
    def type(self):
        pass

    @property
    @abstractmethod
    def full_displayname(self):
        pass

    @property
    def signature(self):
        """Unique identifier of cursor"""
        return self.full_displayname

    @property
    @abstractmethod
    def kind_name(self):
        pass

    @property
    def children(self):
        return self._children

    def add_child(self, node):
        """Add child node to parent node"""
        node.parent = self
        self._children.append(node)

        # if signature is empty (e.g. for unexposed types) we assume the node to be not searchable/reusable
        # existence of unexposed cursor types can be dependent on platform (we have many of them on linux)
        if node.signature:
            # keep the node with template_type_name (type name without template parameters) for lookups
            if node.type == NodeType.CLANG_NODE and node.is_class_or_struct and node.cxx_element.is_templated:
                template_type_name = cutil.template_type_name(node.signature)
                if template_type_name in node.root._node_map:
                    # a template specialization, add it to template nodes map
                    template_node = node.root._node_map[template_type_name]
                    template_node._template_spec_node_map[node.signature] = node
                else:
                    # a template node add it to root node_map
                    node.root._node_map[cutil.template_type_name(node.signature)] = node
            else:
                node.root._node_map[node.signature] = node

    @cached_property
    @allowed_after_build
    def ancestor_with_api(self):
        node = self.parent
        while node:
            if node.api:
                return node
            node = node.parent
        return node

    def walk_preorder(self):

        # processor current node
        yield self

        # now if needed dive into children
        for child in self:
            for descendant in child.walk_preorder():
                yield descendant


class DirectoryNode(Node):

    def __init__(self, name, file_name=None, api=None, args=None, root=None, parent=None, children=None,
                 pure_comment=None):
        super().__init__(api, args, root, parent, children, pure_comment)
        self.name = name
        self._file_name = file_name

    def __iter__(self):
        # directory nodes are skipped from further iteration
        # root provides iteration on files by their processing order
        return self

    def __next__(self):
        # directory nodes are skipped from further iteration
        # root provides iteration on files by their processing order
        raise StopIteration

    @property
    def file_name(self):
        return self._file_name

    @property
    def line_number(self):
        return None

    @property
    def type(self):
        return NodeType.DIRECTORY_NODE

    @property
    def kind_name(self):
        return DIR_KIND_NAME

    @property
    def displayname(self):
        return self.name

    @property
    def full_displayname(self):
        return self.name


class RootNode(Node):
    ROOT_KEY = '__root__'

    def __init__(self):
        super().__init__(api=Node.API_NONE, args=None, root=self, parent=None, children=None, pure_comment=None)
        self.name = RootNode.ROOT_KEY
        self._node_map = {}
        self.__is_ir_built = False
        # IR keeps the order of processed files to provide an iterator with correct order
        self.__ordered_nodes = []

    def __repr__(self):
        return f"RootNode({self.__dict__})"

    def __iter__(self):
        # root node is an iterator and keeps all directories and files by their processing order
        self.idx = 0
        self.__length = len(self.__ordered_nodes)
        return self

    def __next__(self):
        # root node is an iterator and keeps all directories and files by their processing order
        # directories are iterated first and then files in their processing order
        if self.idx < self.__length:
            next_node = self.__ordered_nodes[self.idx]
            self.idx += 1
            return next_node
        raise StopIteration

    def walk(self):
        for node in self.walk_preorder():
            yield node

    def find_node(self, node_signature, match_template=True):
        """Returns node in IR by its unique signature"""
        node = self._node_map.get(node_signature)
        if node:
            return node
        if cutil.is_templated(node_signature):
            template_node = self._node_map.get(cutil.template_type_name(node_signature))
            if template_node:
                return template_node._template_spec_node_map.get(node_signature, None if match_template else template_node)
        return None

    def find_node_by_type(self, type_, match_template=True):
        """Find node by cursor Type, CXXType or string"""

        search_name = type_
        if isinstance(type_, (cli.Type, CXXType)):
            # decl is None for the types which are invalid or unexposed
            decl = cutil.get_declaration(type_) if isinstance(type_, cli.Type) else type_.get_declaration()
            return self.find_node(cutil.get_signature(decl), match_template) if decl else None
        return self.find_node(search_name, match_template)

    def _get_all_nodes(self):
        """Returns map of all nodes"""
        return self._node_map.values()

    def _is_built(self):
        """Protected method to check whether IR is built or not"""
        return self.__is_ir_built

    def _finalize_ir_build(self, processed_files):
        """Protected method to set IR built flag to true and to define ir nodes by correct order."""
        self.__is_ir_built = True
        dir_nodes = []
        self.__get_directory_nodes(self, dir_nodes)
        self.__ordered_nodes = dir_nodes + [self._node_map[file] for file in processed_files]
        self.__length = len(self.__ordered_nodes)

    def __get_directory_nodes(self, node, nodes):
        for child in node.children:
            if child.type == NodeType.DIRECTORY_NODE:
                nodes.append(child)
                self.__get_directory_nodes(child, nodes)

    @property
    def type(self):
        return NodeType.ROOT_NODE

    @property
    def full_displayname(self):
        return "Root"

    @property
    def kind_name(self):
        return ROOT_KIND_NAME

    @property
    def displayname(self):
        return self.name


class ClangNode(Node, ABC):

    def __init__(self, clang_cursor, api=None, args=None, root=None, parent=None, children=None, pure_comment=None):
        super().__init__(api, args, root, parent, children, pure_comment)
        self.cxx_element = CXXElement.create_cxx_element(clang_cursor)

    def __iter__(self):
        # clang node is an iterator and gives its child nodes by their processing order
        self.idx = 0
        self.__length = len(self._children)
        return self

    def __next__(self):
        # clang node is an iterator and gives its child nodes by their processing order
        if self.idx < self.__length:
            next_node = self._children[self.idx]
            self.idx += 1
            return next_node
        raise StopIteration

    @property
    def kind(self):
        return self.cxx_element.kind

    @property
    def displayname(self):
        return self.cxx_element.displayname

    @property
    def spelling(self):
        return self.cxx_element.spelling

    @cached_property
    def full_displayname(self):
        return self.cxx_element.get_full_displayname()

    @cached_property
    def signature(self):
        return self.cxx_element.get_signature()

    @property
    def file_name(self):
        return self.cxx_element.extent.start.file.name

    @property
    def line_number(self):
        return self.cxx_element.extent.start.line


class FileNode(Node):

    def __init__(self, file_path, api=None, args=None, root=None, parent=None, children=None, pure_comment=None):
        super().__init__(api, args, root, parent, children, pure_comment)
        self.file_path = file_path

    def __iter__(self):
        self.idx = 0
        self.__length = len(self._children)
        return self

    def __next__(self):
        if self.idx < self.__length:
            next_node = self._children[self.idx]
            self.idx += 1
            return next_node
        raise StopIteration

    @property
    def displayname(self):
        return self.file_path

    @property
    def spelling(self):
        return self.file_path

    @property
    def line_number(self):
        return None

    @cached_property
    def signature(self):
        return self.file_path

    @property
    def file_name(self):
        return self.file_path

    @property
    def type(self):
        return NodeType.FILE_NODE

    @property
    def kind_name(self):
        return FILE_KIND_NAME

    @property
    def full_displayname(self):
        return self.file_path


class CXXNode(ClangNode):

    def __init__(self, clang_cursor, api=None, args=None, root=None, parent=None, children=None, pure_comment=None):
        super().__init__(clang_cursor, api, args, root, parent, children, pure_comment)
        # keep direct descendants in sorted set to have deterministic order of nodes
        self.direct_descendants = SortedSet()
        self._template_spec_node_map = {}

    @property
    def type(self):
        return NodeType.CLANG_NODE

    @property
    def kind_name(self):
        return self.cxx_element.kind_name

    @property
    def is_class_or_struct(self):
        return self.cxx_element.kind in (ElementKind.STRUCT_DECL,
                                         ElementKind.CLASS_DECL,
                                         ElementKind.CLASS_TEMPLATE,
                                         ElementKind.CLASS_TEMPLATE_PARTIAL_SPECIALIZATION)

    @cached_property
    @allowed_after_build
    def _base_type_specifier_nodes(self):
        if not self.is_class_or_struct:
            return None

        base_type_specifier_nodes = []
        for base_specifier in self.cxx_element.get_children():
            if base_specifier.kind == cli.CursorKind.CXX_BASE_SPECIFIER:
                base_type = base_specifier.type
                # lookup with type spelling, then with canonical type name spelling
                # (to support template usages with namespace)
                base_node = self.root.find_node(base_type.spelling, match_template=False)
                if base_node is None:
                    # canonical type spelling is used in case of typedefs
                    base_node = self.root.find_node(base_type.get_canonical().spelling, match_template=False)

                if base_node:
                    base_type_specifier_nodes.append(base_node)

        return base_type_specifier_nodes

    @cached_property
    @allowed_after_build
    def descendants(self):
        """List of all descendants of struct/class node"""

        if not self.is_class_or_struct:
            return None

        descendants = []
        # recursively construct list of descendants
        for direct_desc in self.direct_descendants:
            for node in direct_desc.descendants:
                if node not in descendants:
                    descendants.append(node)

        descendants.extend(self.direct_descendants)

        return descendants

    @allowed_after_build
    def find_adjacent_nodes(self, search_names, search_api=None):
        """Search for adjacent nodes by name/api"""
        return (self.root.find_node_by_type(node.signature) for node in self.parent.children
                if (search_api is None or node.api == search_api) and node.spelling in search_names)

    def find_adjacent_node(self, search_names, search_api=None):
        """Search for an adjacent node by search name/api"""
        return next(self.find_adjacent_nodes(search_names, search_api), None)

    @cached_property
    @available_on(cli.CursorKind.CXX_METHOD,
                  cli.CursorKind.FUNCTION_DECL,
                  cli.CursorKind.CONSTRUCTOR,
                  cli.CursorKind.FUNCTION_TEMPLATE)
    def overloading_index(self):
        """Find overloading index of the given function/method/ctor node"""

        search_api = self.api
        name = self.spelling
        search_names = {name}
        overloads = self.find_adjacent_nodes(search_names, search_api)
        for i, ctx in enumerate(overloads):
            if ctx == self:
                return i

        return None



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.

"""
Module for defining various error classes
"""

from traceback import extract_stack, format_list

from cppbind import logging


class Error:
    """
    Class is used to call various types of errors
    """

    __has_error = False
    __error_limit = -1

    @classmethod
    def set_error_limit(cls, error_limit):
        """Method to set error limit"""
        cls.__error_limit = error_limit

    @classmethod
    def has_error(cls):
        """Method to check whether error has ever occurred"""
        return cls.__has_error

    @classmethod
    def error(cls, msg, file=None, line=None, err_type_name=None):
        """Simple error type"""
        cls.__has_error = True

        if cls.__error_limit == 0:
            Error.critical(msg, file, line)
        else:
            cls.__error_limit -= 1
            logging.error(cls.get_message(msg, file, line, err_type_name or "error"))
            cls.log_traceback(extract_stack())

    @classmethod
    def critical(cls, msg, file=None, line=None, err_type_name=None):
        """Critical error: it's not possible to proceed after this type of error"""
        error_msg = cls.get_message(msg, file, line, err_type_name or "critical")
        logging.critical(error_msg)
        cls.log_traceback(extract_stack())
        raise CppBindError(error_msg)

    @classmethod
    def warning(cls, msg, file=None, line=None, err_type_name=None):
        """Warning: execution of program is not stopped"""
        logging.warning(cls.get_message(msg, file, line, err_type_name or "warning"))

    @classmethod
    def internal(cls, msg, file=None, line=None, err_type_name=None):
        """Method to report internal errors"""
        Error.critical(msg, file, line, err_type_name or "internal")

    @classmethod
    def get_message(cls, msg, file, line, err_type):
        """Get error message by its type"""
        err_msg = err_type.upper() + ': '
        if file:
            err_msg += f"file -> {file}, "
            if line:
                err_msg += f"line -> {line}, "
        err_msg += f"msg -> {msg}"

        return err_msg

    @classmethod
    def log_traceback(cls, traceback_stack):
        """Get exception traceback and log it in case of debug"""
        if logging.getLevelName(logging.root.level) == "DEBUG":
            traceback_list = format_list(traceback_stack)
            traceback_list.pop()
            logging.debug(''.join(traceback_list))


class CppBindError(Exception):
    """
    Exception class to raise errors in case of critical cppbind errors
    """



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.

from types import SimpleNamespace

from cached_property import cached_property

import cppbind.utils.clang as cutil
from cppbind.ir import ElementKind, TypeKind
from cppbind.ir.cxx_type import CXXType


class CXXExposedType:
    """
    Type which holds a CXXType or a string type with its cursor and template choice if it's a template type,
    and gives some utility functionality based on it.
    """

    def __init__(self, cxx_type, template_choice=None, cxx_element=None):
        # CXXType object or the string representation of the type
        self._cxx_type = cxx_type
        self._template_choice = template_choice
        # if the type is invalid for templates but the cursor is available use it to retrieve information
        self._cxx_element = cxx_element

        # TODO: replace False with the logic for strings when it's possible
        self.is_enum = cxx_type.is_enum if isinstance(cxx_type, CXXType) else False
        self.is_bool = cxx_type.is_bool if isinstance(cxx_type, CXXType) else False
        self.is_long = cxx_type.is_long if isinstance(cxx_type, CXXType) else False
        self.is_float = cxx_type.is_float if isinstance(cxx_type, CXXType) else False
        self.is_char = cxx_type.is_char if isinstance(cxx_type, CXXType) else False
        self.is_uchar = cxx_type.is_uchar if isinstance(cxx_type, CXXType) else False

    def __eq__(self, other):
        return self.type_name == other.type_name

    def __hash__(self):
        # TODO: note this might be required to be changed in #239
        return hash(self.type_name)

    @property
    def type_name(self):
        return cutil.replace_template_choice(
            self._cxx_type if isinstance(self._cxx_type, str) else self._cxx_type.type_name,
            self._template_choice).rstrip('.')

    @property
    def pointee_type(self):
        if isinstance(self._cxx_type, CXXType):
            pointee_type = self._cxx_type.pointee_type
            return CXXExposedType(pointee_type,
                                  self._template_choice) if pointee_type.spelling else self
        # type_ is a string
        _type = self._cxx_type.strip()
        if _type.endswith('&&'):
            return CXXExposedType(_type[:-2].strip(), self._template_choice)
        if _type.endswith('*') or _type.endswith('&'):
            return CXXExposedType(_type[:-1].strip(), self._template_choice)
        return self

    @property
    def canonical_type(self):
        if isinstance(self._cxx_type, CXXType):
            canonical_type = self._cxx_type.canonical_type
            return CXXExposedType(canonical_type,
                                  self._template_choice) if canonical_type.spelling else self
        else:
            # we do not have a mechanism for string types yet
            return self

    @property
    def is_template(self):
        """
        Returns whether the type is a template or not. Only most nested type is considered.
        For example for cppbind::example::List<int>::Item it'll return False as Item itself is not a template.
        This property is used for internal purposes like for building type converters.
        """
        # we have to use type_name for clang types as well as it can be an unexposed type and it's choice can be a
        # template for example if it's spelling is T and the choice of T is std::vector<int> then T is a template
        if isinstance(self._cxx_type, CXXType) and self._cxx_type.kind != TypeKind.UNEXPOSED:
            return self._cxx_type.get_num_template_arguments() != -1
        else:
            return self.type_name.endswith('>')

    @property
    def is_function_proto(self):
        if isinstance(self._cxx_type, CXXType):
            return self._cxx_type.is_function_proto
        # we don't have a mechanism for string types yet
        return False

    @property
    def is_typedef(self):
        if isinstance(self._cxx_type, CXXType):
            return self._cxx_type.is_typedef
        # we don't have  a mechanism for string types yet
        return False

    @cached_property
    def __template_parameter_cursors(self):
        if self._cxx_element:
            return self._cxx_element.get_template_parameter_cursors()
        else:
            cursor = self._cxx_type.get_template_declaration()
            return cutil.get_template_parameter_cursors(cursor)

    @cached_property
    def is_variadic(self):
        """Returns whether the type is a variadic template or an instance of a variadic template."""
        if isinstance(self._cxx_type, str):
            # in case of strings we cannot identify whether it's variadic or not
            return False
        elif isinstance(self._cxx_type, CXXType):
            return self._cxx_type.is_variadic
        else:
            return any(cutil.is_variadic(cursor) for cursor in self.__template_parameter_cursors)

    @cached_property
    def template_arguments(self):
        """
        Returns a list of namespaces containing the argument and its kind, e.g. for std::array<double, 3> returns
        [SimpleNamespace(type=CXXExposedType('double'), kind=CursorKind.TEMPLATE_TYPE_PARAMETER),
         SimpleNamespace(type=3, kind=CursorKind.TEMPLATE_NON_TYPE_PARAMETER)].
        Currently, type and integral parameters are supported.
        """

        if not self.is_template:
            return []

        if self._cxx_element is None and (isinstance(self._cxx_type, str) or self._cxx_type.kind == TypeKind.UNEXPOSED):
            # neither cursor nor type is available, thus using string parsing to retrieve template arguments
            return self.__arguments_from_type_name()

        # either cursor or type is available, thus we can use them to retrieve information about template parameters
        cursors = self.__template_parameter_cursors

        if self.is_variadic:
            # for variadic use string parse
            # NOTE: later when adding support for other template parameter
            # kinds like enums, floating point parameters, etc., it might be reasonable to not use string parsing
            return self.__arguments_from_type_name()

        if isinstance(self._cxx_type, CXXType) and self._cxx_type.kind != TypeKind.UNEXPOSED:
            # clang Type is available, and it's not unexposed, we can take template arguments from it
            arguments = [self._cxx_type.get_template_argument_type(num) for num in
                         range(self._cxx_type.get_num_template_arguments())]
        else:
            # clang Type is not available, but we have a cursor
            # which can be used to retrieve information about template parameters
            arguments = [cursor.spelling for cursor in cursors]

        # argument spellings are used for non type template parameters, as for them clang gives an invalid Type
        # using canonical as the argument might be an expression that should be evaluated
        argument_spellings = cutil.get_template_arguments_from_str(
            self.type_name if self._is_unexposed else self.canonical_type.type_name)
        args = []
        for ii, argument in enumerate(arguments):
            cursor = cursors[ii]
            # for now using clangs CursorKind for identifying type and non-type parameters
            # later we might have our own template parameter kind and expose it to snippets
            if cursor.kind == ElementKind.TEMPLATE_TYPE_PARAMETER:
                args.append(SimpleNamespace(type=CXXExposedType(argument, self._template_choice),
                                            kind=ElementKind.TEMPLATE_TYPE_PARAMETER))
            elif cursor.kind == ElementKind.TEMPLATE_NON_TYPE_PARAMETER:
                if isinstance(argument, CXXType) and argument.kind == TypeKind.INVALID:
                    # if the non-type argument is a clang Type then it's invalid so use its spelling
                    # retrieved from string parsing
                    argument = argument_spellings[ii]
                if cutil.is_integral_type(cursor.type.get_canonical()):
                    args.append(
                        SimpleNamespace(type=int(cutil.replace_template_choice(argument, self._template_choice)),
                                        kind=ElementKind.TEMPLATE_NON_TYPE_PARAMETER))
                else:
                    # non-integral non type parameter, currently not supported
                    args.append(None)
            else:
                # template template parameter, currently not supported
                args.append(None)

        return args

    def __arguments_from_type_name(self):
        args = []
        for arg_spelling in cutil.get_template_arguments_from_str(self.type_name):
            parameter = cutil.replace_template_choice(arg_spelling, self._template_choice)
            if parameter.lstrip('+-').isnumeric():
                args.append(SimpleNamespace(type=int(parameter),
                                            kind=ElementKind.TEMPLATE_NON_TYPE_PARAMETER))
            else:
                # NOTE: currently only from the string we cannot identify whether the argument is a type or not
                # maybe we can check by trying to build its converter?
                args.append(SimpleNamespace(type=CXXExposedType(arg_spelling, self._template_choice),
                                            kind=ElementKind.TEMPLATE_TYPE_PARAMETER))
        return args

    @property
    def _template_type_name(self):
        return cutil.template_type_name(self.type_name)

    @property
    def is_lval_reference(self):
        return self._cxx_type.is_lval_reference if isinstance(self._cxx_type, CXXType) else \
            not self._cxx_type.strip().endswith('&&') and self._cxx_type.strip().endswith('&')

    @property
    def is_rval_reference(self):
        return self._cxx_type.is_rval_reference if isinstance(self._cxx_type, CXXType) else \
            self._cxx_type.strip().endswith('&&')

    @property
    def is_pointer(self):
        return self._cxx_type.is_pointer if isinstance(self._cxx_type, CXXType) else self._cxx_type.strip().endswith(
            '*')

    @property
    def is_value_type(self):
        return not self.canonical_type.is_pointer and not self.canonical_type.is_lval_reference

    @property
    def unqualified_type_name(self):
        return cutil.get_unqualified_type_name(self.type_name).rstrip('.')

    @property
    def resolved_type(self):
        """Returns pointee type if the type is not typedef on pointer, otherwise returns canonical type"""
        return self.canonical_pointee_type if (isinstance(self._cxx_type, CXXType) and self.is_typedef and (
                self.canonical_type.is_pointer or self.canonical_type.is_lval_reference)) else self.pointee_type

    @property
    def is_const_qualified(self):
        # when self._cxx_type is string then only prefixed const qualified types are supported
        # for example `int const &` is not supported
        # in case of clang's Type all cases are supported
        return self.type_name.startswith('const')

    @property
    def base_types(self):
        """
        Returns a list of base exposed types.
        """
        # redefining to add a template choice
        return [] if isinstance(self._cxx_type, str) else [CXXExposedType(base_type, self._template_choice) for
                                                           base_type in self._cxx_type.base_types]

    @property
    def __pointee_name(self):
        return cutil.replace_template_choice(
            self.pointee_type._cxx_type.spelling if isinstance(self._cxx_type,
                                                               CXXType) else self.pointee_type._cxx_type,
            self._template_choice)

    @property
    def canonical_pointee_type(self):
        t = self
        if self.is_typedef or not self._is_unexposed:
            t = self.canonical_type
        if t.is_pointer or t.is_lval_reference:
            t = t.pointee_type
        return t

    @property
    def _is_unexposed(self):
        """
        Recursively checks if the type has an unexposed template argument.
        E.g. std::vector<std::shared_ptr<T>> is unexposed.
        Returns:
            bool: True if the has an unexposed template argument and False otherwise.
        """
        if isinstance(self._cxx_type, str):
            return self._cxx_type != cutil.replace_template_choice(self._cxx_type, self._template_choice)
        clang_type = self.canonical_type._cxx_type
        if clang_type.kind == TypeKind.UNEXPOSED:
            return True
        if clang_type.kind in (TypeKind.POINTER, TypeKind.LVALUEREFERENCE):
            return self.pointee_type._is_unexposed
        if self.is_template:
            for num in range(self._cxx_type.get_num_template_arguments()):
                tp = self._cxx_type.get_template_argument_type(num)
                if tp.kind == TypeKind.UNEXPOSED:
                    return True
        return False


class CXXRunnerExposedType(CXXExposedType):
    """Subclass of CXXExposedType which defines some additional context dependent properties"""

    def __init__(self, cxx_type, template_choice=None, cxx_element=None, ir=None, ctx=None):
        super(CXXRunnerExposedType, self).__init__(cxx_type, template_choice, cxx_element)
        self.__ir = ir
        self.__ctx = ctx

    @cached_property
    def __node(self):
        """Private helper property to find node from IR or context"""

        # first lookup in IR (we may lookup only in IR, when we explode template nodes in IR)
        if self.__ir:
            node = self.__ir.find_node_by_type(self._cxx_type)
            if node:
                return node

        # use context if IR lookup failed
        if self.__ctx:
            return self.__ctx.node

        return None

    @property
    def namespace(self):
        """The namespace of the type"""
        return self.__node.cxx_element.namespace if self.__node else None

    @property
    def is_polymorphic(self):
        """Checks whether the type is polymorphic"""
        return self.__node.cxx_element.is_polymorphic if self.__node else None

    @property
    def parent_type(self):
        """Lexical parent type of the current type"""

        if not self.__node:
            return None

        parent_element = self.__node.cxx_element.parent
        if not parent_element:
            return None

        parent_cxx_type = parent_element.type if not parent_element.is_templated else cutil.replace_template_choice(
            parent_element.get_full_displayname(), self._template_choice or self.__ctx.template_choice)

        return CXXExposedType(parent_cxx_type, self._template_choice)



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.

"""
Parser module based on clang
"""
import glob
import os
import re

import clang.cindex as cli
import cppbind.utils.clang as cutil
from cppbind import logging
from cppbind.common.error import Error


class CXXParser:
    """
    Class is responsible for source cxx files parsing and cursor returning
    """
    CLANG_DEF_OPTIONS = cli.TranslationUnit.PARSE_SKIP_FUNCTION_BODIES | \
                        cli.TranslationUnit.PARSE_INCOMPLETE
    EXTRA_INCLUDES_FILE_NAME = "extra_includes.cpp"

    def __init__(self, processor, single_tu, filter_=None):
        self.current_file = None
        self.filter = filter_
        self.processor = processor
        self.single_tu = single_tu
        # list to keep track of which files are already processed to not process them again
        self.processed_files = []
        # current files stack which are being processed
        self.files_stack = []

    def parse_tu(self):
        """
        Parses cxx files and returns list of TranslationUnit s
        """
        return list(self.parse_x())

    def parse_tu_x(self, clang_args, include_dirs, src_glob, extra_headers, **kwargs):
        """
        Parses cxx files and returns generator of TranslationUnit s
        """
        index = cli.Index.create()
        # build parser arguments
        args = ['-x', 'c++', '--std=c++20'] + clang_args + ['-I' + includeDir.strip()
                                                            for includeDir in
                                                            include_dirs]

        logging.debug(f"Clang path: {cutil.get_libclang_full_path()}")
        logging.debug(f"Clang args: {args}")

        logging.info("parsing files: {}".format(' '.join(src_glob)))

        # using list to keep files order constant
        all_files = []
        for file in src_glob:
            files_glob = sorted(glob.glob(file.strip(), recursive=True))
            for file_path in files_glob:
                abs_fp = os.path.abspath(file_path)
                if not (self.filter and self.filter.filter_by_file(abs_fp)):
                    all_files.append(abs_fp)

        extra_includes = None
        unsaved_files = None
        if extra_headers:
            extra_includes = "\n".join(f"#include <{header}>" for header in extra_headers)
            unsaved_files = [(CXXParser.EXTRA_INCLUDES_FILE_NAME, extra_includes)]

        if self.single_tu and all_files:
            # include all src files in a virtual file to parse everything once
            includes = "\n".join(f'#include "{file}"' for file in all_files)
            if extra_includes:
                includes = extra_includes + f"\n{includes}"
            unsaved_files = [(CXXParser.EXTRA_INCLUDES_FILE_NAME, includes)]
            all_files = [CXXParser.EXTRA_INCLUDES_FILE_NAME]
        else:
            if extra_includes:
                all_files = [CXXParser.EXTRA_INCLUDES_FILE_NAME] + all_files

        for file in all_files:
            try:
                tu = index.parse(path=file,
                                 args=args,
                                 unsaved_files=unsaved_files,
                                 options=CXXParser.CLANG_DEF_OPTIONS)
            except cli.TranslationUnitLoadError as err:
                Error.critical(f"Cannot continue: cppbind error has occurred: {err}.")
            has_error = False
            for diagnostic in tu.diagnostics:
                diagnostic_error_msg = diagnostic.spelling
                if diagnostic.severity in (cli.Diagnostic.Error, cli.Diagnostic.Fatal):
                    Error.error(diagnostic_error_msg,
                                   diagnostic.location.file,
                                   diagnostic.location.line)
                    has_error = True
                else:
                    Error.warning(diagnostic_error_msg,
                                  diagnostic.location.file,
                                  diagnostic.location.line)
            if not has_error and tu.spelling not in self.processed_files:
                    # allow different files(includes) in translation unit level
                    self._process_cursor_children(tu.cursor, allow_includes=True)
                    if self.files_stack:
                        for _ in range(len(self.files_stack)):
                            self._end_process_file(self.files_stack[-1])

    def cursor_walk(self, cursor):
        """
        cursor_walk ast recursively by filtering using filter
        """
        if self.filter and self.filter.filter_cursor(cursor):
            logging.debug(f"Filtering cursor: {cursor.spelling}")
            return

        # processor current cursor
        yield cursor

        # now if needed dive into children
        if not (self.filter and self.filter.filter_cursor_children(cursor)):
            for child in cursor.get_children():
                for descendant in self.cursor_walk(child):
                    yield descendant

    def _start_process_file(self, file_name):
        logging.info(f'parsing {file_name}')
        self.files_stack.append(file_name)
        tu_parent_dirs = CXXParser.__dirs_to_process(file_name)
        # TODO:
        # currently we are not parsing the entire directory tree at once,
        # instead for each file we are going from root to leaf,
        # if there are already processed dirs then we are retrieving them from the dict,
        # later we may construct the whole source directory tree and then build the ir tree
        for dir_name in tu_parent_dirs:
            if hasattr(self.processor, 'start_dir'):
                self.processor.start_dir(dir_name)

        if hasattr(self.processor, 'start_file'):
            self.processor.start_file(file_name)

    def _end_process_file(self, file_name):
        tu_parent_dirs = CXXParser.__dirs_to_process(file_name)

        if hasattr(self.processor, 'end_file'):
            self.processor.end_file(file_name)

        if tu_parent_dirs:
            for dir_name in reversed(tu_parent_dirs):
                if hasattr(self.processor, 'end_dir'):
                    self.processor.end_dir(dir_name)

        if not self.single_tu:
            self.processed_files.append(file_name)
        self.files_stack.pop()

    def parse(self, **kwargs):
        """
        Method parses all translation units
        """
        self.parse_tu_x(**kwargs)

    @staticmethod
    def __dirs_to_process(file_name):
        if not os.path.isfile(file_name):
            return []

        dirs_to_search = set()

        root = os.path.abspath(file_name)
        while root:
            root = os.path.dirname(root)
            dir_name = os.path.relpath(root, os.getcwd())

            match = re.match(r'^[./]+$', dir_name) and dir_name != '.'
            if match:
                break
            dirs_to_search.add(dir_name)
        # sort to get root to child list
        return sorted(dirs_to_search)

    def _process_cursor(self, cursor, allow_includes=False):
        if self.filter and self.filter.filter_cursor(cursor):
            return

        # process current cursor
        if hasattr(self.processor, 'start_cursor'):
            self.processor.start_cursor(cursor)

        if callable(self.processor):
            self.processor(cursor)

        self._process_cursor_children(cursor, allow_includes)

        if hasattr(self.processor, 'end_cursor'):
            self.processor.end_cursor(cursor)

    def _process_cursor_children(self, cursor, allow_includes=False):
        current_allow_includes = cursor.kind not in (
            cli.CursorKind.NAMESPACE,
            cli.CursorKind.STRUCT_DECL,
            cli.CursorKind.CLASS_DECL,
            cli.CursorKind.CLASS_TEMPLATE,
            cli.CursorKind.CLASS_TEMPLATE_PARTIAL_SPECIALIZATION,
            cli.CursorKind.ENUM_DECL,
            cli.CursorKind.FUNCTION_DECL,
            cli.CursorKind.FUNCTION_TEMPLATE)
        for child in cursor.get_children():
            process_child = child.kind in (
                cli.CursorKind.NAMESPACE,
                cli.CursorKind.STRUCT_DECL,
                cli.CursorKind.CLASS_DECL,
                cli.CursorKind.CLASS_TEMPLATE,
                cli.CursorKind.CLASS_TEMPLATE_PARTIAL_SPECIALIZATION,
                cli.CursorKind.CONSTRUCTOR,
                cli.CursorKind.CXX_METHOD,
                cli.CursorKind.FIELD_DECL,
                cli.CursorKind.ENUM_DECL,
                cli.CursorKind.FUNCTION_DECL,
                cli.CursorKind.FUNCTION_TEMPLATE)
            if process_child:
                if child.extent.start.file:
                    file = os.path.abspath(child.extent.start.file.name)
                    # file is changed, either we ignore it if it's not allowed(e.g. include in namespace)
                    # or start/continue processing the new/not_finished file
                    current_file = self.files_stack[-1] if self.files_stack else None
                    if file != current_file:
                        # consider parent allow includes and current allow includes
                        if allow_includes and current_allow_includes:
                            # file is already started to be processed but due to includes is not finalized yet, e.g.
                            # void func1()
                            # #include "other.h"
                            # void func2()
                            if file in self.files_stack:
                                child_file = self.files_stack[-1]
                                # end included files to come to current file to be able to continue its processing
                                while file != child_file:
                                    self._end_process_file(child_file)
                                    child_file = self.files_stack[-1]
                                self._process_cursor(child, allow_includes and current_allow_includes)
                            else:
                                # file is not processed, start its processing
                                # if merge src is true we can not have duplicate files
                                if self.single_tu or file not in self.processed_files:
                                    self._start_process_file(file)
                                    self._process_cursor(child, allow_includes and current_allow_includes)
                    else:
                        # file is not changed, continue processing
                        self._process_cursor(child, allow_includes and current_allow_includes)
            else:
                self._process_cursor_children(child, allow_includes and current_allow_includes)



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.

"""
This module is responsible for loading, processing code, types, actions snippets
"""

import copy
import filecmp
import glob
import os
import shutil
from collections.abc import MutableMapping
from functools import lru_cache, partial
from operator import attrgetter
from types import SimpleNamespace
from typing import Iterable

from jinja2 import TemplateSyntaxError, Template

from cppbind import BANNER_LOGO, logging
from cppbind.common import JINJA_UNIQUE_MARKER
from cppbind.common.error import Error, CppBindError
from cppbind.common.type_info import create_type_info
from cppbind.common.yaml_process import to_value
from cppbind.cxx_exposed import CXXExposedType
from cppbind.ir import ElementKind
from cppbind.utils import (
    get_public_attributes,
    init_jinja_env,
)

OBJECT_INFO_TYPE = '$Object'
ENUM_INFO_TYPE = '$Enum'
FUNCTION_PROTO_INFO_TYPE = '$FunctionProto'

# snippets
FILE = 'file'
FILE_PATH = 'file_path'
CONTENT = 'content'
UNIQUE_CONTENT = 'unique_content'
SCOPES = 'scopes'

# actions
FILES_GLOB = 'files_glob'
FILES_ACTION_KIND = 'file'
FILE_ACTION_VARIABLE = 'variables'
COPY_ACTION = 'copy_to'
RENDER_ACTION = 'render_to'


class Snippet:
    """
    Class represents a snippet with template and own context
    """

    def __init__(self, context, template, marker=None):
        self.context = context
        self.template = template
        self.marker = marker

    def __str__(self):
        value = self.template.render(self.context)
        if self.marker:
            return value.replace(self.marker, '')
        return value


class JinjaTemplate:
    def __init__(self, jinja_env, template_str, scope, file, line_number):
        try:
            self.template = jinja_env.from_string(template_str) if isinstance(template_str, str) else template_str
        except TemplateSyntaxError as e:
            Error.critical(f"Jinja syntax error in {scope} - {str(e)}" if scope else f"Jinja syntax error - {str(e)}",
                           file=file,
                           line=line_number + e.lineno)
        self.scope = scope
        self.file = file
        self.line_number = line_number

    def render(self, context):
        try:
            return self.template.render(context) if isinstance(self.template, Template) else self.template
        except CppBindError:
            raise
        except Exception as e:
            raise Error.critical(
                f'Jinja render error in {self.scope} - {str(e)}.', file=self.file, line=self.line_number)


class Action:
    """Base class for any action subclass"""

    def __init__(self, kind_name):
        self.kind_name = kind_name

    def do(self, context):
        return {}


class FileAction(Action):
    """Class represents file actions"""

    def __init__(self, jinja_env, src_glob_tmpls, dest_tmpl, variables_tmpl):
        super().__init__(FILES_ACTION_KIND)
        self.src_glob_tmpls = src_glob_tmpls
        self.dest_tmpl = dest_tmpl
        self.variables_tmpl = variables_tmpl
        self.jinja_env = jinja_env

    def do(self, ctx):
        variables = {name: [] for name in self.variables_tmpl}
        src_globs = [tmpl.render(ctx) for tmpl in self.src_glob_tmpls]

        def _make_context(ctx):
            def make():
                # helper variables
                path = os.path
                pat_sep = os.sep
                file_name = None
                banner_logo = BANNER_LOGO
                return locals()

            context = copy.copy(ctx)
            context.update(make())
            return context

        context = _make_context(ctx)

        for src_file in [fl for gl in src_globs for fl in sorted(glob.glob(gl, recursive=True))]:
            context['file_name'] = src_file
            if self.dest_tmpl:
                target_file = self.dest_tmpl.render(context)
                if target_file:
                    try:
                        self.do_main_action(src_file, target_file, context)
                    except OSError as err:
                        Error.error(f"CppBind couldn't do action: {err}")

            # update variables
            for var_name, tmpl in self.variables_tmpl.items():
                variables[var_name].append(tmpl.render(context))

        return variables


class CopyFileAction(FileAction):
    """Class represents file copy action"""

    def do_main_action(self, src_file, target_file, context):
        """Implementation logic of file copy action"""

        os.makedirs(os.path.dirname(target_file), exist_ok=True)
        if not os.path.isfile(target_file) or not filecmp.cmp(src_file, target_file):
            shutil.copyfile(src_file, target_file)


class RenderFileAction(CopyFileAction):
    """Class represents template file render and copy action"""

    def do_main_action(self, src_file, target_file, context):
        """Implementation logic of file render action"""

        os.makedirs(os.path.dirname(target_file), exist_ok=True)
        src_file_content = JinjaTemplate(self.jinja_env,
                                         open(src_file).read(),
                                         None,
                                         os.path.abspath(src_file),
                                         0).render(context)
        target_file_content = None
        if os.path.isfile(target_file):
            target_file_content = open(target_file).read()

        if target_file_content != src_file_content:
            with open(target_file, 'w') as f:
                f.write(src_file_content)


class TypeConverter:
    """Base type converter class"""

    def __init__(self,
                 type_info,
                 type_info_collector,
                 template_args,
                 **kwargs):
        self.type_info = type_info
        self.__type_info_collector = type_info_collector
        self.__template_args = template_args
        self.__kwargs = kwargs
        self.__context = self.__make_context()
        self.custom = self.__eval_custom(self.__context)
        self.__create_section_converters()

    @property
    def template_args(self):
        """List of template argument types"""
        return [arg.type for arg in self.__template_args]

    @property
    def template_args_kinds(self):
        """List of template argument kinds"""
        return [arg.kind for arg in self.__template_args]

    def __make_context(self):

        def make():
            # helper variables
            template_args = self.template_args
            template_args_kinds = self.template_args_kinds

            cxx = self.type_info.cxx
            vars = self.type_info.vars
            has_api = self.type_info.has_api
            template_args_postfixes = self.type_info.template_args_postfixes
            descendants = self.type_info.descendants

            return locals()

        context = make()
        del context['self']

        # expose cxx and vars so that they can be overridden if required when calling snippets
        self.__expose_type_namespace(context, 'cxx')
        self.__expose_namespace(context, 'vars')

        return context

    def __eval_custom(self, context):
        custom = SimpleNamespace()
        # evaluate custom fields one by one to make available by defined order
        for k, v in self.__type_info_collector.custom.__dict__.items():
            setattr(custom, k, v.render(context))
            # to make accessible for coming fields
            context['custom'] = custom

        return custom

    def __expose_namespace(self, context, name):
        if context[name]:
            for k, v in context[name].__dict__.items():
                context[f'{name}_{k}'] = v

    def __expose_type_namespace(self, context, name):
        if context[name]:
            for prop in get_public_attributes(CXXExposedType.__dict__):
                context[f'cxx_{prop}'] = getattr(context[name], prop)

    def __create_section_converters(self):
        """Creates concrete section type converters"""

        for name in self.__type_info_collector.target_type_infos:
            type_converter_info = self.__type_info_collector.target_type_infos[name]
            type_section_converter = TypeSectionConverter(type_converter_info, self.__context)
            setattr(self, name, type_section_converter)

        for name in self.__type_info_collector.converters:
            type_converter_info = self.__type_info_collector.converters[name]
            type_section_converter = TypeSectionConverter(type_converter_info, self.__context)
            setattr(self, name, type_section_converter)


class TypeSectionConverter:
    """Concrete type section converter"""

    def __init__(self, type_converter_info, context):
        self.__type_converter_info = type_converter_info
        self.__context = context

    def snippet(self, name=None, **kwargs):
        return self.__type_converter_info.snippet(name, {**self.__context, **kwargs})

    def converted_name(self, name):
        return self.__type_converter_info.converted_name(name)

    @property
    def target_type_name(self):
        return self.__type_converter_info.target_type_name(self.__context)

    def get_target_type_name(self, **kwargs):
        return self.__type_converter_info.target_type_name(self.__context, **kwargs)

    @property
    def source_type_name(self):
        return self.__type_converter_info.source_type_name(self.__context)


class TypeInfoCollector:

    def __init__(self, name, target_type_infos, converters, custom, **kwargs):
        self.name = name
        self.converters = converters
        self.target_type_infos = target_type_infos
        self.custom = custom
        self.kwargs = kwargs

    def make_type_converter(self, type_info, template_args):
        return TypeConverter(type_info=type_info,
                             type_info_collector=self,
                             template_args=template_args,
                             **self.kwargs)


class TargetTypeInfo:

    def __init__(self, name, target_type_info):
        self.target_type_info = target_type_info
        self.name = name

    def target_type_name(self, context, **kwargs):
        if self.target_type_info is None:
            return None
        return self.target_type_info.render({**context, **kwargs})


class TypeConvertorInfo(TargetTypeInfo):

    def __init__(self, snippet_tmpl, source_type_info, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.source_type_info = source_type_info
        self.snippet_tmpl = snippet_tmpl
        self._converted_prefix = self.name.replace('_', '')

    def source_type_name(self, context):
        if self.source_type_info is None:
            return None
        return self.source_type_info.render(context)

    def snippet(self, name, context):
        if self.snippet_tmpl:
            return self.snippet_tmpl.render({'name': name,
                                             'target_name': self.converted_name(name),
                                             'target_type_name': self.target_type_name(context),
                                             'get_target_type_name': partial(self.target_type_name, context),
                                             **context})
        return ""

    def converted_name(self, name):
        return f"{self._converted_prefix}{name}" if self.snippet_tmpl else name


class ScopeInfo:

    def __init__(self, name, parent_scopes, scopes, snippet_tmpl, unique_snippet_tmpl):
        self.name = name
        self.parent_scopes = parent_scopes
        self.scopes = scopes
        self.snippet_tmpl = snippet_tmpl
        self.unique_snippet_tmpl = unique_snippet_tmpl

    def make_snippet(self, context, marker=JINJA_UNIQUE_MARKER):
        return Snippet(context=context, template=self.snippet_tmpl, marker=marker)

    def unique_make_snippet(self, context, marker=None):
        return Snippet(context=context, template=self.unique_snippet_tmpl, marker=marker)


class FileScopeInfo(ScopeInfo):

    def __init__(self, file_path_tmpl, parent_scopes, scopes, snippet_tmpl, unique_snippet_tmpl):
        super().__init__(name=FILE, parent_scopes=parent_scopes,
                         scopes=scopes, snippet_tmpl=snippet_tmpl,
                         unique_snippet_tmpl=unique_snippet_tmpl)
        self.file_path_tmpl = file_path_tmpl

    def get_file_name(self, context):
        return self.file_path_tmpl.render(context)


class SnippetsEngine:
    TARGET_KEY = 'target'
    SOURCE_KEY = 'source'
    SNIPPET_KEY = 'snippet'
    CUSTOM_KEY = 'custom'
    TYPES_KEY = 'types'
    CONVERTERS_KEY = 'converters'

    def __init__(self, runner, ctx_desc, platform, language):
        self.runner = runner
        self.platform = platform
        self.language = language
        self.ctx_desc = ctx_desc
        self.type_infos = {}
        self.file_infos = {}
        self.code_infos = {}
        self.action_infos = []
        self.jinja_env = init_jinja_env(language)
        self.__update_jinja_env()

    def load(self):
        action_snippets = self.ctx_desc.get_action_snippets()
        code_snippets = self.ctx_desc.get_code_snippets()
        type_converter_snippets = self.ctx_desc.get_type_converter_snippets()

        if action_snippets:
            self._load_actions(action_snippets[self.language][self.platform])
        if code_snippets:
            self._load_code_info(code_snippets[self.language][self.platform])
        if type_converter_snippets:
            self._load_type_info(type_converter_snippets[self.language][self.platform])

    def do_actions(self, context):
        variables = {}
        for act in self.action_infos:
            vars_ = act.do(context)
            variables.update(vars_)
        return variables

    @lru_cache(maxsize=512)
    def build_type_converter(self, cxx_exposed_type):

        res = self._build_type_converter(cxx_exposed_type)

        if res is None:
            raise KeyError(f"Can not find type for {cxx_exposed_type.type_name}")

        return res

    def get_type_info(self, type_name):
        return self.type_infos.get(type_name, None)

    def get_code_info(self, code_name):
        return self.code_infos.get(code_name, None)

    def get_file_info(self, file_name):
        return self.file_infos.get(file_name)

    def _remove_placeholder_vars(self, info_dict):
        variables = [key for key in info_dict if key.startswith('var')]
        for key in variables:
            del info_dict[key]

    def _load_actions(self, actions_info):
        def handle_file_action(info_dict):
            src_glob_tmpls = src_glob_node = info_dict[FILES_GLOB]

            if not src_glob_tmpls.is_of_type(list):
                src_glob_tmpls = [src_glob_tmpls.value]
            src_glob_tmpls = [JinjaTemplate(self.jinja_env,
                                            tmpl,
                                            f'{FILES_ACTION_KIND}:{FILES_GLOB}',
                                            src_glob_node.file,
                                            src_glob_node.line_number) for tmpl in src_glob_tmpls]

            variables = info_dict.get('variables', {})
            variables_tmpl = {}
            for var_name, var_tmpl in variables.items():
                var_tmpl = JinjaTemplate(self.jinja_env,
                                         var_tmpl.value,
                                         f'{FILES_ACTION_KIND}:{FILE_ACTION_VARIABLE}:{var_name}',
                                         var_tmpl.file,
                                         var_tmpl.line_number)
                variables_tmpl[var_name] = var_tmpl

            if COPY_ACTION in info_dict and info_dict[COPY_ACTION]:
                action_node = info_dict[COPY_ACTION]
                self.action_infos.append(CopyFileAction(self.jinja_env,
                                                        src_glob_tmpls,
                                                        JinjaTemplate(
                                                            self.jinja_env,
                                                            action_node.value,
                                                            f'{FILES_ACTION_KIND}:{COPY_ACTION}',
                                                            action_node.file,
                                                            action_node.line_number),
                                                        variables_tmpl))
            elif RENDER_ACTION in info_dict and info_dict[RENDER_ACTION]:
                action_node = info_dict[RENDER_ACTION]
                self.action_infos.append(RenderFileAction(self.jinja_env,
                                                          src_glob_tmpls,
                                                          JinjaTemplate(self.jinja_env,
                                                                        action_node.value,
                                                                        f'{FILES_ACTION_KIND}:{RENDER_ACTION}',
                                                                        action_node.file,
                                                                        action_node.line_number),
                                                          variables_tmpl))

        def handle_default(info_dict):
            Error.critical("undefined actions.")

        actions_map = {
            'file': handle_file_action
        }
        # load into structures
        for action_info in actions_info:
            for action_name, action_data in action_info.items():
                actions_map.get(action_name, handle_default)(action_data)

    def _load_code_info(self, code_info_dict):
        # load into structures
        self._load_file_info(code_info_dict[FILE])
        del code_info_dict[FILE]
        # remove variables since they are used for placeholders
        self._remove_placeholder_vars(code_info_dict)

        for code_name, info_map in code_info_dict.items():
            if info_map.is_of_type(str):
                # redirection
                info_map = code_info_dict[info_map.value]

            if info_map.value is None:
                # allow empty rules
                continue

            if not isinstance(info_map, MutableMapping):
                Error.critical("Missing scopes section.")

            self.code_infos[code_name] = self._load_code_structure_info(code_name, info_map)

    def _load_code_structure_info(self, code_name, code_info_dict):
        def scope_walk(info_map, scopes=None):
            scopes = scopes or tuple()
            if not info_map.is_of_type(MutableMapping):
                # leaf node
                yield scopes, SimpleNamespace(content=info_map, unique_content=None, scopes=[])
            elif CONTENT in info_map or UNIQUE_CONTENT in info_map:
                yield (scopes, SimpleNamespace(content=info_map.get(CONTENT, None),
                                               unique_content=info_map.get(UNIQUE_CONTENT, None),
                                               scopes=info_map.get(SCOPES, [])))
            else:
                for scope, child_info_map in info_map.items():
                    for d in scope_walk(child_info_map, scopes + tuple([scope])):
                        yield d

        scope_infos = {}

        for scopes, info in scope_walk(code_info_dict):
            snippet = info.content
            if snippet:
                snippet = JinjaTemplate(self.jinja_env,
                                        snippet.value,
                                        f'{code_name}:{":".join(scopes)}:{CONTENT}',
                                        snippet.file,
                                        snippet.line_number)

            unique_snippet = info.unique_content
            if unique_snippet:
                unique_snippet = JinjaTemplate(self.jinja_env,
                                               unique_snippet.value,
                                               f"{code_name}:{':'.join(scopes)}:{UNIQUE_CONTENT}",
                                               unique_snippet.file,
                                               unique_snippet.line_number)

            scope_infos[scopes] = ScopeInfo(name=code_name,
                                            parent_scopes=scopes,
                                            scopes=info.scopes,
                                            snippet_tmpl=snippet,
                                            unique_snippet_tmpl=unique_snippet)

        return scope_infos

    def _load_file_info(self, code_info_dict):
        # load into structures
        for file_name, info_map in code_info_dict.items():
            file_node = info_map[FILE_PATH]
            file_path = JinjaTemplate(self.jinja_env,
                                      file_node.value,
                                      f'{FILE}:{file_name}:{FILE_PATH}',
                                      file_node.file,
                                      file_node.line_number)
            snippet = info_map.get(CONTENT, None)
            if snippet:
                snippet = JinjaTemplate(self.jinja_env,
                                        snippet.value,
                                        f'{FILE}:{file_name}:{CONTENT}',
                                        snippet.file,
                                        snippet.line_number)
            unique_snippet = info_map.get(UNIQUE_CONTENT, None)
            if unique_snippet:
                unique_snippet = JinjaTemplate(self.jinja_env,
                                               unique_snippet.value,
                                               f'{FILE}:{file_name}:{UNIQUE_CONTENT}',
                                               unique_snippet.file,
                                               unique_snippet.line_number)
            self.file_infos[file_name] = FileScopeInfo(file_path,
                                                       [file_name],
                                                       info_map.get(SCOPES, []),
                                                       snippet,
                                                       unique_snippet)

    def _load_type_info(self, type_info_dict):
        # remove variables since they are used for placeholders
        self._remove_placeholder_vars(type_info_dict)
        # load into structures
        for type_name, info_map in type_info_dict.items():
            if info_map.is_of_type(str):
                # redirection
                info_map = type_info_dict[info_map.value]

            if not isinstance(info_map, MutableMapping):
                Error.critical("Missing type information")

            custom = SimpleNamespace()
            target_types = {}
            type_converters = {}

            # iterate over all sections: 'custom', 'types', 'converters'
            for section_key, section_val in info_map.items():
                if section_key == SnippetsEngine.CONVERTERS_KEY:
                    # converter
                    for name, info in section_val.items():
                        target_lang = source_lang = None
                        if '_to_' in name:
                            index = name.rfind('_to_')
                            target_lang = name[index + 4:]
                            source_lang = name[:index]

                        if info.is_of_type(dict):
                            target_lang = to_value(info.get(SnippetsEngine.TARGET_KEY, target_lang))
                            source_lang = to_value(info.get(SnippetsEngine.SOURCE_KEY, source_lang))
                        target_type_info = source_type_info = None
                        if target_lang and source_lang:
                            target_tmpl = source_tmpl = '{{cxx.type_name}}'
                            file = section_val.file
                            # setting the initial line number the start of the converter
                            # i.e. the line number of cxx type name
                            source_line_number = target_line_number = info_map.line_number
                            if target_lang in info_map[SnippetsEngine.TYPES_KEY]:
                                target_node = info_map[SnippetsEngine.TYPES_KEY][target_lang]
                                target_tmpl = target_node.value
                                target_line_number = target_node.line_number
                            if source_lang in info_map[SnippetsEngine.TYPES_KEY]:
                                source_node = info_map[SnippetsEngine.TYPES_KEY][source_lang]
                                source_tmpl = source_node.value
                                source_line_number = source_node.line_number

                            if target_tmpl is not None and source_tmpl is not None:
                                target_type_info = JinjaTemplate(self.jinja_env,
                                                                 target_tmpl,
                                                                 f'{type_name}:{SnippetsEngine.TYPES_KEY}:{target_lang}',
                                                                 file,
                                                                 target_line_number)
                                source_type_info = JinjaTemplate(self.jinja_env,
                                                                 source_tmpl,
                                                                 f'{type_name}:{SnippetsEngine.TYPES_KEY}:{source_lang}',
                                                                 file,
                                                                 source_line_number)

                        snippet_tmpl = info and JinjaTemplate(self.jinja_env,
                                                              info[SnippetsEngine.SNIPPET_KEY].value if info.is_of_type(
                                                                  dict) else info.value,
                                                              f'{type_name}:{SnippetsEngine.CONVERTERS_KEY}:{name}',
                                                              info.file,
                                                              info.line_number)

                        target_info = TypeConvertorInfo(snippet_tmpl=snippet_tmpl,
                                                        name=name,
                                                        target_type_info=target_type_info,
                                                        source_type_info=source_type_info)
                        type_converters[name] = target_info
                elif section_key == SnippetsEngine.CUSTOM_KEY:
                    # primitive type info
                    custom = {}
                    for k, v in section_val.items():
                        custom[k] = JinjaTemplate(
                            self.jinja_env,
                            v.value,
                            f'{type_name}:custom:{k}',
                            v.file,
                            v.line_number)

                    custom = SimpleNamespace(**custom)
                else:
                    # type info
                    for name, info in section_val.items():
                        target_type_info = JinjaTemplate(self.jinja_env,
                                                         info.value,
                                                         f'{type_name}:{SnippetsEngine.TYPES_KEY}:{name}',
                                                         info.file,
                                                         info.line_number)
                        target_info = TargetTypeInfo(name=name, target_type_info=target_type_info)
                        target_types[name] = target_info

            self.type_infos[type_name] = TypeInfoCollector(name=type_name,
                                                           target_type_infos=target_types,
                                                           converters=type_converters,
                                                           custom=custom)

    def build_type_converter_with_typename(self, type_name, template_choice=None):
        return self.build_type_converter(CXXExposedType(type_name, template_choice))

    def _create_type_info(self, search_name, cxx_exposed_type, template_args=None, **kwargs):
        logging.debug(f"Finding type for {search_name}")
        ref_ctx = self.runner.get_context(search_name)
        if ref_ctx is not None:
            if ref_ctx.node.kind == ElementKind.ENUM_DECL:
                search_name = ENUM_INFO_TYPE
            else:
                search_name = OBJECT_INFO_TYPE

        type_converter = self.get_type_info(search_name)

        if type_converter is None:
            # type info for given type is not available
            return None

        type_info = create_type_info(self.runner, cxx_exposed_type)

        template_args = [arg for arg in template_args if arg] if template_args else []
        type_converter = type_converter.make_type_converter(type_info, template_args)

        return type_converter

    def _build_type_converter(self,
                              cxx_exposed_type: CXXExposedType,
                              lookup_type: CXXExposedType = None,
                              search_name: str = None):
        template_choice = cxx_exposed_type._template_choice or {}

        lookup_type = lookup_type or cxx_exposed_type

        search_name = search_name or lookup_type.type_name
        search_name = template_choice.get(search_name, search_name)
        logging.debug(f"Creating type converter for {search_name} "
                      f"and template choice {template_choice}")
        type_info = None
        if not lookup_type.is_template or self.get_type_info(search_name) is not None:
            # `self.get_type_info(search_name) is not None` is to allow specialized template converters
            type_info = self._create_type_info(search_name,
                                               cxx_exposed_type=cxx_exposed_type)

        if type_info is None:
            unqualified_name = lookup_type.unqualified_type_name
            if unqualified_name != search_name:
                return self._build_type_converter(cxx_exposed_type,
                                                  lookup_type,
                                                  search_name=unqualified_name)
            pointee_type = lookup_type.pointee_type
            if pointee_type != lookup_type:
                return self._build_type_converter(cxx_exposed_type,
                                                  pointee_type)
            if lookup_type.is_function_proto:
                # here we assume lookup_type holds clang type
                # update later when we will be able to parse lambda type from string

                # TODO, maybe not only ElementKind.TEMPLATE_TYPE_PARAMETER
                tmpl_args = [SimpleNamespace(type=CXXExposedType(arg_type, cxx_exposed_type._template_choice),
                                             kind=ElementKind.TEMPLATE_TYPE_PARAMETER)
                             for arg_type in lookup_type._cxx_type.argument_types]
                tmpl_args.append(SimpleNamespace(type=CXXExposedType(lookup_type._cxx_type.get_result(),
                                                                     cxx_exposed_type._template_choice),
                                                 kind=ElementKind.TEMPLATE_TYPE_PARAMETER))

                type_info = self._create_type_info(FUNCTION_PROTO_INFO_TYPE,
                                                   cxx_exposed_type=cxx_exposed_type,
                                                   template_args=tmpl_args)
                return type_info

            # covers template parameter and template argument cases,
            # e.g. a::Stack<T> and a::Stack<Project>
            # might be a template typedef so get the canonical type and then proceed
            if lookup_type.is_template and not lookup_type.is_typedef:
                # currently, only integral parameter is supported from non-type parameters
                tmpl_args = lookup_type.template_arguments

                # for the case when all arguments are exposed
                # for example a::Stack<Project> then the canonical will
                # return type with spelling equal to a::Stack<b::Project>
                # this won´t work if there's an unexposed argument e.g.T,
                # for example for the case a::Stack<T>, the canonical
                # will remove namespaces and return
                # type with spelling equal to 'Stack<type-parameter-0-0>'
                if not lookup_type._is_unexposed:
                    lookup_type = lookup_type.canonical_type

                # at first search with full exposed name
                type_info = self._create_type_info(lookup_type.type_name,
                                                   cxx_exposed_type=cxx_exposed_type,
                                                   template_args=tmpl_args)
                if not type_info:
                    # if not found with full name search without arguments
                    type_info = self._create_type_info(lookup_type._template_type_name,
                                                       cxx_exposed_type=cxx_exposed_type,
                                                       template_args=tmpl_args)
                return type_info

            canonical_type = lookup_type.canonical_type
            if canonical_type != lookup_type:
                return self._build_type_converter(cxx_exposed_type, canonical_type)

        return type_info

    def __create_type_converter(self, type_, error=True):
        """Function to create a converter for a given type"""

        try:
            if isinstance(type_, str):
                return self.build_type_converter_with_typename(type_)
            return self.build_type_converter(type_)
        except KeyError:
            if error:
                raise
            return None

    def __get_type_info(self, type_, error=True):
        """Function to return the type info for a given type"""

        converter = self.__create_type_converter(type_, error=error)
        return converter.type_info if converter else None

    def __create_types_converters(self, input_, attribute=None, error=True):
        """Function to create converter(s) for given type(s)"""

        if isinstance(input_, Iterable) and not isinstance(input_, str):
            return [self.__create_types_converters(item, attribute, error) for item in input_]

        try:
            input_value = attrgetter(attribute)(input_) if attribute else input_
        except AttributeError as err:
            Error.critical(f"Error while applying 'type_converter' filter: {err}")
        else:
            if not isinstance(input_value, (str, CXXExposedType)):
                Error.critical(f"'type_converter' filter is applicable only for 'str' and 'CXXExposedType' types "
                               f"(or a list of those types). {type(input_value)} is an unexpected type.")

            return self.__create_type_converter(input_value, error)

    def __get_types_infos(self, input_, attribute=None, error=True):
        """Function to return the type info(s) for given type(s)"""

        if isinstance(input_, Iterable) and not isinstance(input_, str):
            return [self.__get_types_infos(item, attribute, error) for item in input_]

        try:
            input_value = attrgetter(attribute)(input_) if attribute else input_
        except AttributeError as err:
            Error.critical(f"Error while applying 'type_info' filter: {err}")
        else:
            if not isinstance(input_value, (str, CXXExposedType)):
                Error.critical(f"'type_info' filter is applicable only for 'str' and 'CXXExposedType' types "
                               f"(or a list of those types). {type(input_value)} is an unexpected type.")

            return self.__get_type_info(input_value, error)

    def __has_api(self, cxx_element):
        return self.runner.get_context(cxx_element._get_full_displayname()) is not None

    def __update_jinja_env(self):
        """Add custom filters/globals to jinja environment"""

        self.jinja_env.filters['type_converter'] = self.__create_types_converters
        self.jinja_env.filters['type_info'] = self.__get_types_infos

        self.jinja_env.globals['get_type_converter'] = self.__create_type_converter
        self.jinja_env.globals['get_type_info'] = self.__get_type_info

        self.jinja_env.globals['has_api'] = self.__has_api
        self.jinja_env.tests['has_api'] = self.__has_api
        self.jinja_env.filters['has_api'] = self.__has_api



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.

"""
Helper codes for kotlin conversion
"""
from . import *


OPERATOR_MAPPING = {
    '*': 'times',
    '-': 'minus',
    '+': 'plus',
    '/': 'div',
    '==': 'equals',
    '%': 'rem',
    '!': 'not',
    '()': 'invoke',
    '[]': 'get',
}

JNI_TYPE_SIGNATURE = dict(
    jboolean='Z',
    jbyte='B',
    jchar='C',
    jshort='S',
    jint='I',
    jlong='J',
    jfloat='F',
    jdouble='D',
    jobjectid='J'
)


def get_operator_name(spelling):
    operator = spelling.replace('operator', '').strip()
    return OPERATOR_MAPPING.get(operator, spelling)


def get_jni_type_signature(native_type):
    return JNI_TYPE_SIGNATURE.get(native_type)


def arg_str(type_name, name, default=None, **kwargs):
    arg_str = name + ': ' + type_name
    if default:
        val = default
        if val in ['nullptr', 'NULL']:
            arg_str += '? = null'
        else:
            arg_str += " = " + val
    return arg_str


def get_jni_func_name(package_name, class_name, method_name, args_type_name=None):
    def fix_name(name):
        for s, r in [('_', '_1'), ('.', '_'), (';', '_2'), ('[', '_3')]:
            name = name.replace(s, r)
        return name

    args_type_signature = dict(
        jboolean='Z',
        jbyte='B',
        jchar='C',
        jshort='S',
        jint='I',
        jlong='J',
        jfloat='F',
        jdouble='D',
        jobject='Ljava_lang_Object_2',
        jstring='Ljava_lang_String_2',
    )
    package_name = fix_name(package_name)
    # consider nested types as well
    class_name = '_00024'.join([fix_name(t) for t in class_name.split('.')])
    method_name = fix_name(method_name)
    if args_type_name is None or any((a not in args_type_signature for a in args_type_name)):
        return f'Java_{package_name}_{class_name}_{method_name}'
    return f'Java_{package_name}_{class_name}_{method_name}__\
{"".join([args_type_signature[arg] for arg in args_type_name])}'



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.

import re
import uuid

from cppbind import DATETIME_REGEX

OUTPUT_MODIFICATION_KEY = str(uuid.uuid1())


def is_output_changed(old, new):
    """
    Compares cppbind generated old code with new one.
    It's assumed that to each occurrence of OUTPUT_MODIFICATION_KEY
    in new code corresponds a datetime in old one.
    Args:
        old(str): CppBind generated old code.
        new(str): CppBind generated new code.
    Returns:
        bool: Returns true if old and new do not match and false otherwise.
    """
    keyword_len = len(OUTPUT_MODIFICATION_KEY)
    idx = new.find(OUTPUT_MODIFICATION_KEY)
    while idx != -1:
        new_to = new[:idx]
        old_to = old[:idx]
        if new_to != old_to:
            # compare until keyword occurrence
            return True
        # drop compared part with keyword
        new = new[idx + keyword_len:]
        old = old[idx:]
        # check old starts with datetime
        match = re.match(DATETIME_REGEX, old)
        if match is None:
            return True

        # drop datetime from old
        old = old[match.span()[1]:]
        idx = new.find(OUTPUT_MODIFICATION_KEY)
    # compare remaining part
    return old != new



"""
  ______ .______   .______   .______    __  .__   __.  _______
 /      ||   _  \  |   _  \  |   _  \  |  | |  \ |  | |       \
|  ,----'|  |_)  | |  |_)  | |  |_)  | |  | |   \|  | |  .--.  |
|  |     |   ___/  |   ___/  |   _  <  |  | |  . `  | |  |  |  |
|  `----.|  |      |  |      |  |_)  | |  | |  |\   | |  '--'  |
 \______|| _|      | _|      |______/  |__| |__| \__| |_______/

This file is generated by cppbind on 06/23/2021-10:23.
Please do not change it manually.
"""

import functools
import importlib
import inspect

__all__ = ['CppBindMetaclass', 'CppBindEnumMetaclass']


class CppBindBaseMetaclass(type):
    """
    Base class for CppBindMetaclass and CppBindEnumMetaclass.
    """

    def __instancecheck__(self, instance):
        """
        This is responsible for correct instance check for cppbind class.
        As for wrapper's an instance of pybind class is returned then we need this to return true for wrapper instances.
        """
        # get pybind's corresponding type
        pybind_class = _load_pybind_type(self.__name__, self.__module__)
        # check if wrapper instance is instance of pybind type
        return isinstance(instance, pybind_class)


class CppBindMetaclass(CppBindBaseMetaclass):
    """
    Metaclass for all cppbind generated classes. It's responsible for replacing pybind's methods/properties with
    corresponding cppbind's methods/properties.
    Overrides __new__ and __init__ to create an instance of pybind class instead of cppbind's and changes
    base types for the types which are extended from cppbind types.
    """

    def __new__(mcs, future_class_name, future_class_parents, future_class_attrs, *args, **kwargs):
        """
        For user extended types it replaces base types which are cppbind generated types with original pybind types
        and overwrites cppbind's metaclass with pybind's metaclass.
        """
        pybind_type = _load_pybind_type(future_class_name, future_class_attrs['__module__'])
        # set metaclass for extended types to pybind metaclass
        meta_class = None
        if pybind_type is None:
            parents = []
            for parent in future_class_parents:
                if isinstance(parent, CppBindMetaclass):
                    # get pybind's corresponding type
                    pybind_class = _load_pybind_type(parent.__name__, parent.__dict__['__module__'])
                    parents.append(pybind_class)
                    if meta_class is None:
                        # set pybind metaclass for non cppbind generated types
                        meta_class = type(pybind_class)
                else:
                    parents.append(parent)
            future_class_parents = tuple(parents)
        meta_class = meta_class or mcs
        return super(CppBindMetaclass, mcs).__new__(meta_class, future_class_name, future_class_parents,
                                                    future_class_attrs, **kwargs)

    def __init__(cls, future_class_name, future_class_parents, future_class_attrs):
        """
        Replaces pybind's method/property with cppbind's method/property and attaches pybind's method/property as an
        attribute to cppbind method. Also overwrites __new__ for cppbind types to create an instance of pybind class.
        """
        # get pybind's corresponding type
        pybind_class = _load_pybind_type(future_class_name, future_class_attrs['__module__'])
        if pybind_class:
            originals = pybind_class.__dict__.copy()

            for attr in pybind_class.__dict__:
                if attr in future_class_attrs and attr != '__new__':
                    # add original pybind method/property to cppbind method/property as an attribute
                    if hasattr(future_class_attrs[attr], 'fn'):
                        # method decorated with @bind
                        setattr(future_class_attrs[attr].fn.original_function, attr, originals[attr])
                    elif inspect.isfunction(future_class_attrs[attr]):
                        # not decorated method
                        setattr(future_class_attrs[attr], attr, originals[attr])
                    elif isinstance(future_class_attrs[attr], property):
                        # properties
                        bound_getter = future_class_attrs[attr].fget
                        if bound_getter:
                            setattr(bound_getter.fn.original_function, attr, originals[attr].fget)
                        bound_setter = future_class_attrs[attr].fset
                        if bound_setter:
                            setattr(bound_setter.fn.original_function, attr, originals[attr].fset)
                    elif isinstance(future_class_attrs[attr], classmethod):
                        # static functions
                        bound_method = future_class_attrs[attr].__func__
                        setattr(bound_method.fn.original_function, attr,
                                originals[attr].__get__(future_class_attrs[attr]))

                    # replace pybind method with cppbind method
                    setattr(pybind_class, attr, future_class_attrs[attr])
            # there's no nesting in pybind using this to support nested types
            setattr(pybind_class, '__qualname__', future_class_attrs['__qualname__'])
            # overwrite __new__ to return create a pybind instance instead of cppbind instance
            setattr(cls, '__new__', functools.partial(_new_object, pybind_class))
        super(CppBindMetaclass, cls).__init__(future_class_name, future_class_parents, future_class_attrs)


class CppBindEnumMetaclass(CppBindBaseMetaclass):
    """
    Metaclass for all cppbind generated enums. It's responsible for adding user defined attributes to pybind enum type.
    Also overwrites cppbind enum's __new__ to create a pybind enum.
    """

    def __init__(cls, future_class_name, future_class_parents, future_class_attrs):
        """
        Adds cppbind generated enum attributes to corresponding pybind enum.
        """
        # get pybind's corresponding type
        pybind_enum = _load_pybind_type(future_class_name, future_class_attrs['__module__'])
        for attr in pybind_enum.__dict__:
            if attr in future_class_attrs and attr not in ('__doc__', '__qualname__', '__module__', '__new__'):
                # replace pybind method with wrapper method
                setattr(pybind_enum, attr, future_class_attrs[attr])
        # overwrite __new__ to create a pybind enum
        setattr(cls, '__new__', functools.partial(_new_object, pybind_enum))
        super(CppBindEnumMetaclass, cls).__init__(future_class_name, future_class_parents, future_class_attrs)


def _new_object(pybind_type, cls, *args, **kwargs):
    """Helper used to overwrite cppbind __new__ to create a pybind type."""
    return pybind_type(*args, **kwargs)


def _load_pybind_type(type_name, cppbind_module_name):
    """Loads and returns corresponding pybind module if it's available, otherwise None."""
    module = importlib.import_module(cppbind_module_name)
    file_name = cppbind_module_name.split('.')[-1]
    # cppbind adds an import of corresponding pybind module in cppbind module with convention = pybind_{cppbind_module_name}
    # then using this import loads pybind type
    pybind_module = getattr(module, f'pybind_{file_name}', None)
    if hasattr(pybind_module, type_name):
        return getattr(pybind_module, type_name)
    return None



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.


"""
Module to define which cxx properties must be exposed to jinja snippets.
"""
import itertools
import types
from copy import copy

from cached_property import cached_property

import cppbind.utils.clang as cutil
from cppbind.cxx_exposed.cxx_exposed_type import CXXExposedType
from cppbind.ir import ElementKind
from cppbind.ir.cxx_element import CXXArgumentElement


class CXXExposedElement:
    """Class to define exposed properties to snippets"""

    def __init__(self, cxx_element, template_choice=None):
        self._cxx_element = cxx_element
        self._template_choice = template_choice

        self.kind_name = cxx_element.kind_name
        self.name = cxx_element.name
        self.displayname = cxx_element.displayname
        self.source_file_name = cxx_element.source_file_name
        self.is_public = cxx_element.is_public
        self.is_private = cxx_element.is_private
        self.is_protected = cxx_element.is_protected
        self.raw_comment = cxx_element.raw_comment
        self.namespace = cxx_element.namespace
        self.is_abstract = cxx_element.is_abstract
        self.is_open = cxx_element.is_open

    def _get_full_displayname(self):
        return self._cxx_element.get_full_displayname()

    @cached_property
    def parent(self):
        return CXXExposedElement.__create_cxx_exposed_element(self._cxx_element.parent, self._template_choice)

    @staticmethod
    def __create_cxx_exposed_element(cxx_element, template_choice):
        """Static factory method to create a cxx exposed element based on the cxx element kind"""

        kind = cxx_element.kind
        if kind == ElementKind.ENUM_DECL:
            return CXXEnumExposedElement(cxx_element)
        if kind in (
                ElementKind.FUNCTION_DECL,
                ElementKind.FUNCTION_TEMPLATE,
                ElementKind.CXX_METHOD,
                ElementKind.CONSTRUCTOR
        ):
            return CXXFunctionExposedElement(cxx_element, template_choice)
        if kind in (
                ElementKind.CLASS_DECL,
                ElementKind.STRUCT_DECL,
                ElementKind.CLASS_TEMPLATE,
                ElementKind.CLASS_TEMPLATE_PARTIAL_SPECIALIZATION
        ):
            return CXXClassExposedElement(cxx_element, template_choice)
        if kind == ElementKind.PARM_DECL:
            return CXXArgumentExposedElement(cxx_element, template_choice)
        if kind == ElementKind.FIELD_DECL:
            return CXXMemberExposedElement(cxx_element, template_choice)

        return CXXExposedElement(cxx_element, template_choice)


class CXXEnumExposedElement(CXXExposedElement):
    """Class to define exposed enum properties to snippets"""

    def __init__(self, cxx_element):
        super(CXXEnumExposedElement, self).__init__(cxx_element)

        self.type_name = cxx_element.type_name
        self.enum_cases = cxx_element.enum_cases


class CXXClassExposedElement(CXXExposedElement):
    """Class to define exposed class/struct properties to snippets"""

    def __init__(self, cxx_element, template_choice=None):
        super(CXXClassExposedElement, self).__init__(cxx_element, template_choice)
        self.__is_templated = cxx_element.is_templated

        self.is_polymorphic = cxx_element.is_polymorphic
        self.has_multiple_base_branches = cxx_element.has_multiple_base_branches
        self.is_template = cxx_element.is_template
        self.template_parameters = cxx_element.template_parameters

    @cached_property
    def type_name(self):
        """The type name corresponding to the given cursor"""

        # in case of templates type.spelling is empty or partial defined (without template parameter specialization)
        # always using display name for unexposed types e.g T
        cxx_type_name = self._cxx_element.get_full_displayname()
        return cutil.replace_template_choice(cxx_type_name, self._template_choice)

    @cached_property
    def base_types(self):
        """
        Returns:
            List of base types exposed elements with template choice.
        """
        # specializing this in exposed type due to template_choice
        return [CXXExposedType(base, self._template_choice) for base in self._cxx_element.base_types]

    @cached_property
    def ancestors(self):
        """
        Returns:
            List of base types exposed elements with template choice.
        """
        # specializing this in exposed type due to template_choice
        return [CXXExposedType(ancestor.type, self._template_choice) for ancestor in
                self._cxx_element.ancestors]

    @cached_property
    def type(self):
        return CXXExposedType(self.type_name if self._cxx_element.is_templated else self._cxx_element.type,
                              template_choice=self._template_choice)


class CXXFunctionExposedElement(CXXExposedElement):
    """Class to define exposed function/method properties to snippets"""

    def __init__(self, cxx_element, template_choice):
        super(CXXFunctionExposedElement, self).__init__(cxx_element, template_choice)

        self.is_template = cxx_element.is_template
        self.is_static = cxx_element.is_static
        self.is_const = cxx_element.is_const
        self.is_overloaded = cxx_element.is_overloaded

    @cached_property
    def overridden_elements(self):
        return [CXXFunctionExposedElement(element, template_choice=self._template_choice)
                for element in self._cxx_element.overridden_elements]

    @cached_property
    def result_type(self):
        return CXXExposedType(self._cxx_element.result_type, template_choice=self._template_choice) \
            if self._cxx_element.result_type else None

    @cached_property
    def args(self):
        _args = []
        template_choice = copy(self._template_choice)
        for arg in self._cxx_element.args:
            # if the argument is a template parameter pack, e.g., Args... then we unpack it
            # and expose each as a separate argument
            # for example if the choice for Args is [int, float]
            # then we expose two arguments corresponding to int and float respectively
            if arg.is_variadic:
                for k in template_choice:
                    if not isinstance(template_choice[k], list):
                        template_choice[k] = [template_choice[k]]
                for i, choice_values in enumerate(list(itertools.product(*template_choice.values()))):
                    # copy the argument and set index to overwrite the name
                    # e.g. first argument - args1, second - args2 and so on
                    arg_copy = CXXArgumentElement(arg._clang_cursor, i)
                    _args.append(CXXArgumentExposedElement(arg_copy,
                                                           template_choice=dict(
                                                               zip(template_choice.keys(), choice_values))))
            else:
                _args.append(CXXArgumentExposedElement(arg, template_choice=self._template_choice))
        return _args

    @cached_property
    def template_arguments(self):
        """Return list of namespaces containing template argument type and kind information."""
        _args = []
        if self.is_template:
            for param in self._cxx_element.template_parameters:
                value = self._template_choice[param.spelling]
                if isinstance(value, list):
                    for item in value:
                        _args.append(types.SimpleNamespace(
                            value=CXXExposedType(item,
                                                 self._template_choice) if param.kind == ElementKind.TEMPLATE_TYPE_PARAMETER else item,
                            kind=param.kind
                        ))
                else:
                    val = CXXExposedType(value,
                                         self._template_choice) if param.kind == ElementKind.TEMPLATE_TYPE_PARAMETER else value
                    _args.append(types.SimpleNamespace(value=val,
                                                       kind=param.kind))
        return _args


class CXXArgumentExposedElement(CXXExposedElement):
    """Class to define exposed argument properties to snippets"""

    def __init__(self, cxx_element, template_choice=None):
        super(CXXArgumentExposedElement, self).__init__(cxx_element, template_choice)

        self.default = cxx_element.default
        self.default_is_literal = cxx_element.default_is_literal
        self.default_is_nullptr = cxx_element.default_is_nullptr

    @cached_property
    def type(self):
        return CXXExposedType(self._cxx_element.type, template_choice=self._template_choice)


class CXXMemberExposedElement(CXXExposedElement):
    """Class to define exposed member properties to snippets"""

    def __init__(self, cxx_element, template_choice=None):
        super(CXXMemberExposedElement, self).__init__(cxx_element, template_choice)

    @cached_property
    def result_type(self):
        return CXXExposedType(self._cxx_element.type, self._template_choice)



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.

"""
Implements cppbind api parser on cxx comment
"""

import re
from collections import OrderedDict
from collections import defaultdict
from types import SimpleNamespace

import yaml
from jinja2.exceptions import TemplateSyntaxError

from cppbind import default_config
from cppbind.common.error import Error
from cppbind.common.yaml_process import has_type, to_value, UniqueKeyLoader
from cppbind.context_manager.var_eval import VariableEvaluator
from cppbind.ir.ast import Node
from cppbind.utils.clang import extract_pure_comment

TEMPLATE_NON_TYPE_PARAMETER_KEY = 'value'
TEMPLATE_TYPE_PARAMETER_KEY = 'type'
TEMPLATE_NAME_KEY = 'name'
TEMPLATE_INSTANCE_KEY = 'template_instance'
TEMPLATE_INSTANCE_ARGS_KEY = 'args'
TEMPLATE_KEY = 'template'


class APIParser:
    """
    A class which is responsible for parsing/validating API annotation section.
    """

    API_START_KW = default_config.api_start_kw
    ALL_PLATFORMS = sorted(default_config.platforms)

    def __init__(self, ctx_desc, jinja_env, platform, language):
        self.ctx_desc = ctx_desc
        self.var_def = self.ctx_desc.get_var_def()
        self.platform = platform
        self.language = language
        self.var_evaluator = VariableEvaluator(jinja_env)

    @staticmethod
    def separate_pure_and_api_comment(raw_comment, index=None):
        """
        A method to separate user defined 'api' related and not related parts
        from the whole pure doxygen comment.
        """
        index = index or raw_comment.find(APIParser.API_START_KW)
        if index == -1:
            return extract_pure_comment(raw_comment), None
        return (extract_pure_comment(raw_comment, index),
                raw_comment[index + len(APIParser.API_START_KW)::])

    def parse_comments(self, api_section, location=None):
        """
        Parse comment to extract API actions and their variables
        """
        if api_section is None:
            return None, OrderedDict()

        # skip comment end
        skip_pattern = r'^\s*(/\*|\*+/|\/+|//!)\s*$'

        # keep empty lines or empty comments
        comment_pattern = r'^\s*(\*|/{3,}|\//!)\s*'

        lines = api_section.splitlines()
        yaml_lines = []

        for line in lines:
            # continue if line contains is only multiline comment begin/end sign
            if re.match(skip_pattern, line):
                continue

            # search for comment prefix
            comment_prefix = re.search(comment_pattern, line)
            if comment_prefix:
                pattern_len = comment_prefix.end()
                yaml_lines.append(' ' * pattern_len + line[pattern_len:])
            else:
                # if no any comment sign found just save the line
                yaml_lines.append(line)

        try:
            attrs = yaml.load('\n'.join(yaml_lines), Loader=UniqueKeyLoader) or {}
            return self.parse_api_attrs(attrs, location)
        except yaml.YAMLError as err:
            Error.critical(f"Error while scanning yaml style comments: {err}",
                           location.file_name if location else None,
                           location.line_number if location else None)

        return None

    def parse_api(self, name, api_section, location, ctx=None):
        """
        Parse api annotations: either comments in source file
        or in separate yaml config files.
        """
        if api_section:
            return self.parse_comments(api_section, location)
        return self.parse_yaml_api(name, ctx, location)

    def parse_yaml_api(self, name, ctx=None, location=None):
        """
        A method to parse api comments from separate yaml file.
        """
        ctx = ctx or {}
        attrs = self.ctx_desc.get_yaml_api(name)
        if attrs:
            # for dir api pass yaml file path
            location = location or SimpleNamespace(file_name=attrs.file, line_number=None)
            try:
                api_attrs = self.var_evaluator.eval_attr_template(attrs, ctx)
                return self.parse_api_attrs(api_attrs, location)
            except TemplateSyntaxError as e:
                Error.critical(
                    f"Jinja syntax error - {str(e)}",
                    file=attrs.file,
                    line=attrs.line_number)
            except Exception as e:
                Error.critical(f"Jinja render error while evaluating yaml api - {e}", attrs.file, attrs.line_number)

        return None

    def parse_api_attrs(self, attrs, location):
        """
        Parsing api comment rows and extracting an action and its corresponding variables.
        """

        curr_lang = self.language
        curr_plat = self.platform

        attr_dict = OrderedDict()
        attr_key_regex = rf"[\s*/]*(?:({'|'.join(APIParser.ALL_PLATFORMS)})\.)?(?:({'|'.join(self.ctx_desc.get_all_languages())})\.)?([^\d\W]\w*)\s*$"

        # Data structure to keep previous priorities
        prev_priors = defaultdict(lambda: [0])
        api = Node.API_NONE
        for attr_key, value in attrs.items():
            match = re.match(attr_key_regex, attr_key)
            if not match:
                Error.critical(f"Wrong variable specification: '{attr_key}'",
                               location.file_name if location else None,
                               location.line_number if location else None)
            platform, language, attr = match.groups()

            if (language and language != curr_lang) or (platform and platform != curr_plat):
                continue

            prior = APIParser.get_priority(platform, language)

            if attr == 'action':
                api = to_value(value)

            if attr not in self.var_def:
                Error.critical(f"Variable {attr} is not specified. "
                               f"It should be one of {set(self.var_def)}.",
                               location.file_name if location else None,
                               location.line_number if location else None)

            curr_max_prior = max(prev_priors[attr])
            # overwrite the value only if the current option
            # has higher priority than all previous ones.
            if prior > curr_max_prior:
                attr_dict[attr] = value
            # If we have this case it means we have a conflict of options: plat.lang and lang.plat
            if prior in prev_priors[attr]:
                Error.error(f"Conflicting variables definition: variables like platform.attr and "
                            f"language.attr cannot be defined together: "
                            f"{curr_lang + '.' + attr, curr_plat + '.' + attr}",
                            location.file_name, location.line_number)
            prev_priors[attr].append(prior)

        return api, attr_dict

    def validate_attr(self, attr_name, attr_value, location=None):
        """Validate attributes"""

        attr_type = self.var_def[attr_name].get('type')
        # TODO: remove these validations when there'll be a mechanism to validate complex variables like list, dict
        if attr_type == 'list':
            if not has_type(attr_value, list):
                Error.critical(f"Wrong variable type: {type(attr_value)}, it must be a list",
                               location.file_name if location else None,
                               location.line_number if location else None)
            if attr_name == TEMPLATE_INSTANCE_KEY:
                for template_instance in attr_value:
                    if not isinstance(template_instance, dict):
                        Error.critical(
                            f"Wrong {TEMPLATE_INSTANCE_KEY} variable style: {attr_value}, "
                            f"items of {TEMPLATE_INSTANCE_KEY} must be of dictionary type.",
                            location.file_name if location else None,
                            location.line_number if location else None
                        )
                    if TEMPLATE_INSTANCE_ARGS_KEY not in template_instance:
                        Error.critical(
                            f"Wrong {TEMPLATE_INSTANCE_KEY} variable style: {attr_value}, "
                            f"{TEMPLATE_INSTANCE_KEY} must have mandatory '{TEMPLATE_INSTANCE_ARGS_KEY}' variable.",
                            location.file_name if location else None,
                            location.line_number if location else None
                        )
        if attr_type == 'dict':
            if not has_type(attr_value, dict):
                Error.critical(f"Wrong variable type: {type(attr_value)}, it must be a dictionary",
                               location.file_name if location else None,
                               location.line_number if location else None)
            if attr_name == TEMPLATE_KEY:
                for attrs in attr_value.values():
                    for attr in attrs:
                        if not isinstance(attr, dict) or TEMPLATE_TYPE_PARAMETER_KEY not in attr \
                                and TEMPLATE_NON_TYPE_PARAMETER_KEY not in attr:
                            Error.critical(
                                f"Wrong template variable style: {attr_value}, "
                                f"template must have mandatory '{TEMPLATE_NON_TYPE_PARAMETER_KEY}'"
                                f" or '{TEMPLATE_TYPE_PARAMETER_KEY}' variable",
                                location.file_name if location else None,
                                location.line_number if location else None
                            )
        # default string type
        return attr_value

    @staticmethod
    def has_api(raw_comment):
        """
        Tests whether the comment has an API section.
        """
        return raw_comment and APIParser.API_START_KW in raw_comment

    @staticmethod
    def get_priority(plat, lang):
        """
        A method to get priority of platform/language specific variable:
        Priorities sorted descending: platform.language.attr, [language|platform].attr, attr
        """
        if plat is None and lang is None:
            return 1
        if plat is None or lang is None:
            return 2
        return 3



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.

from cppbind.common.error import Error
from cppbind.parser import TEMPLATE_TYPE_PARAMETER_KEY, TEMPLATE_NON_TYPE_PARAMETER_KEY

NEW_LINE = '\n'


def make_doxygen_comment(pure_comment):
    if isinstance(pure_comment, str):
        pure_comment = pure_comment.split(NEW_LINE)
    nl = f'{NEW_LINE} * '
    if not pure_comment or all((not line or line.isspace() for line in pure_comment)):
        return ''
    start = '' if not pure_comment[0] or pure_comment[0].isspace() else nl
    return f"""/**{start}{nl.join(pure_comment)}
*/"""


def validate_getter(cxx, vars, setter):
    if cxx.args:
        Error.critical(
            f'Getter should not have arguments: {cxx.parent.displayname}.{cxx.displayname}.')
    if setter:
        if len(setter['cxx'].args) != 1:
            Error.critical(
                f'Setter should have one argument: {cxx.parent.displayname}.{cxx.displayname}.')

        have_diff_nullability = (len(setter['vars'].nullable_arg) == 0) ^ (vars.nullable_return is False)
        if have_diff_nullability:
            Error.critical(
                f'Setter argument and getter return value should have the same nullability:'
                f' {cxx.parent.displayname}.{cxx.displayname}.')


def validate_template_getter_setter(cxx, vars, setter):
    if not setter or not cxx.is_template:
        return

    is_valid = len(vars.template) == len(setter['vars'].template)
    if is_valid:
        for template_arg, possible_types in vars.template.items():
            getter_types = {template.get(TEMPLATE_TYPE_PARAMETER_KEY) or template.get(TEMPLATE_NON_TYPE_PARAMETER_KEY)
                            for template in possible_types}
            setter_types = {template.get(TEMPLATE_TYPE_PARAMETER_KEY) or template.get(TEMPLATE_NON_TYPE_PARAMETER_KEY)
                            for template in setter['vars'].template[template_arg]}
            if getter_types != setter_types:
                is_valid = False
                break
    if not is_valid:
        parent_name = cxx.parent.displayname
        Error.critical(
            f'Template getter/setter should have the same template argument types: '
            f'{parent_name}.{cxx.displayname} and {parent_name}.{setter["cxx"].displayname}.')


def relative_package(namespace1, namespace2):
    ii = 0
    end_idx = 0
    l = min(len(namespace2), len(namespace1))
    while ii < l and namespace1[ii] == namespace2[ii]:
        if namespace1[ii] == '.':
            end_idx = ii + 1
        ii += 1

    return namespace1[end_idx:]


def is_inherited_from_enable_shared_from_this(cxx, directly=False):
    hase_direct_base = any('enable_shared_from_this' in base.type_name for base in cxx.base_types)
    if directly or hase_direct_base:
        return hase_direct_base
    if cxx.base_types:
        return any(is_inherited_from_enable_shared_from_this(base) for base in cxx.base_types)
    return False



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.


from clang.cindex import CursorKind as ElementKind
from clang.cindex import TypeKind, Type



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.

"""
Helper codes for swift conversion
"""
import re

from . import *

DECL_IMPORT_PREFIX_REGEX = r'^(import class |import enum |import func |import struct |import var )'
TEST_IMPORT_PREFIX = '@testable'


def get_c_func_name(prefix, name, postfix):
    new_name = name
    not_allowed = {
        '*': '_mul_',
        '*=': '_imul_',
        '-': '_sub_',
        '-=': '_isub_',
        '+': '_add_',
        '+=': '_iadd_',
        '/': '_truediv_',
        '/=': '_itruediv_',
        '<': '_lt_',
        '<=': '_le_',
        '>': '_gt_',
        '>=': '_ge_',
        '==': '_eq_',
        '!=': '_ne_',
        '%': '_mod_',
        '&': '_and_',
        '|': '_or_',
        '|=': '_ior_',
        '^': '_xor_',
        '^=': '_ixor_',
        '~': '_invert_',
        '~=': '_iinvert_',
        '<<': '_lshift_',
        '>>': '_rshift_',
        '[]': '_getitem_',
    }
    if name in not_allowed:
        new_name = not_allowed[name]
    return f"_func_{prefix}_{new_name}{postfix}"


def imports_comparator(first: str, second: str):
    """
    Comparator for swift imports sorting.
     Swift imports should be in the following order:
        1. Regular imports like 'import ABC'
        2. Declaration imports like 'import func ABC.sort.'
        3. Test modules imports like '@testable import ModuleUnderTest'
    """
    first_is_test = first.startswith(TEST_IMPORT_PREFIX)
    second_is_test = second.startswith(TEST_IMPORT_PREFIX)
    first_is_decl = re.match(DECL_IMPORT_PREFIX_REGEX, first)
    second_is_decl = re.match(DECL_IMPORT_PREFIX_REGEX, second)
    if first_is_test and second_is_test:
        return 1 if first > second else -1
    elif first_is_test:
        return 1
    elif second_is_test:
        return -1
    if first_is_decl and second_is_decl:
        return 1 if first > second else -1
    elif first_is_test:
        return 1
    elif second_is_decl:
        return -1
    else:
        return 1 if first > second else -1


def get_operator_name(name):
    return name[8:] if name.startswith("operator") else name



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.

"""
Filter module decides which clang cursor needs to be processed and which one needs to be skipped.
"""
import os

import clang.cindex as cli
import cppbind.utils.clang as cutil


class CXXFilter:
    """Base class for CXX filter classes"""

    def filter_cursor(self, cursor):
        """Default implementation of `filter_cursor`"""
        return False

    def filter_by_file(self, file):
        """Default implementation of `filter_by_file`"""
        return False

    def filter_cursor_children(self, cursor):
        """Default implementation of `filter_cursor_children`"""
        return False


class CXXComposerFilter(CXXFilter):
    """Class to compose CXXFilers to be able to combine them"""

    def __init__(self, *filters):
        self.__filters = filters

    def filter_cursor(self, cursor):
        """Composer method to call all `filter_cursor` methods of registered filters"""
        return any(filter_.filter_cursor(cursor) for filter_ in self.__filters)

    def filter_by_file(self, file):
        """Composer method to call all `filter_by_file` methods of registered filters"""
        return any(filter_.filter_by_file(file) for filter_ in self.__filters)

    def filter_cursor_children(self, cursor):
        """Composer method to call all `filter_cursor_children` methods of registered filters"""
        return any(filter_.filter_cursor_children(cursor) for filter_ in self.__filters)


class CXXParserFilter(CXXFilter):
    """
    Simple filter implementation to skip unnecessary records for AST of clang.
    """

    def __init__(self, include_files=None, exclude_files=None):
        self.include_files = include_files
        self.exclude_files = exclude_files

    def include_files():
        """
        File list that should be included.
        If this list is not empty all cursors that not belonging
        this files are filtered.

        :returns: list of files
        """
        def fget(self):
            return self._include_files

        def fset(self, value):
            if value is None:
                value = []

            if not isinstance(value, list):
                value = [value]
            self._include_files = value

        return locals()

    include_files = property(**include_files(), doc=include_files.__doc__)

    @staticmethod
    def __has_disallowed_cxx_kind(cursor):
        """Private static method to check whether cursor kind is disallowed for parsing"""
        return cursor.kind.is_reference()

    def filter_cursor(self, cursor):
        """Function test whether or not cursor needs to be processed.

        args:
            cursor: cursor that need to be tested

        Returns:
            True if cursor needs to be processed

        """

        # skip struct/class/enum cursors which are not definitions
        if cursor.kind in [cli.CursorKind.CLASS_DECL, cli.CursorKind.ENUM_DECL,
                           cli.CursorKind.STRUCT_DECL, cli.CursorKind.CLASS_TEMPLATE,
                           cli.CursorKind.CLASS_TEMPLATE_PARTIAL_SPECIALIZATION] and not cursor.is_definition():
            return True

        # skip function(template)/method cursors which are not declarations
        if cursor.kind in [cli.CursorKind.FUNCTION_DECL, cli.CursorKind.FUNCTION_TEMPLATE,
                           cli.CursorKind.CXX_METHOD, cli.CursorKind.CONSTRUCTOR] \
                and cursor.lexical_parent != cursor.semantic_parent:
            return True

        # skip cursors which have disallowed kinds
        if CXXParserFilter.__has_disallowed_cxx_kind(cursor):
            return True

        file = cursor.extent.start.file
        # file is None especially in case of unexposed cursor types
        if file is None or self.filter_by_file(os.path.abspath(file.name)):
            return True

        return False

    def filter_by_file(self, file):
        if self.exclude_files is not None:
            if file in self.exclude_files:
                return True
        if self.include_files:
            return file not in self.include_files
        return False


class CXXOnceProcessingFilter(CXXFilter):
    """Clang cursor filter which has some dependencies on some CppBind internal structures (e.g. IR)"""

    def __init__(self, ir):
        self.__ir = ir

    def filter_cursor_children(self, cursor):
        """
        Function test whether or not cursor children needs to be processed.

        args:
            cursor: cursor that need to be tested
        Returns:
            True if cursor needs to be processed
        """

        # condition to not process cursor children if current cursor-based node is already in node_map
        return self.__ir.find_node(cutil.get_signature(cursor)) is not None


class CppBindFilter(CXXComposerFilter):
    """General purpose cppbind filter"""
    def __init__(self, ir, include_files=None, exclude_files=None):
        super().__init__(CXXParserFilter(include_files, exclude_files),
                         CXXOnceProcessingFilter(ir))



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.

"""
Module is responsible for evaluation of jinja expressions inside variables context definitions.
"""

import yaml
from jinja2.exceptions import TemplateSyntaxError

from cppbind.common.error import Error
from cppbind.common.yaml_process import (
    to_value,
    has_type,
    UniqueKeyLoader,
    YamlNode
)
from cppbind.utils import get_var_real_type


class VariableEvaluator:
    """
    A class to evaluate jinja expressions according to variables type
    """

    def __init__(self, jinja_env):
        self.jinja_env = jinja_env

    def eval_attr_template(self, attrs, ctx):
        """
        Evaluate jinja expressions in current yaml section.
        """

        # if the whole section is pure string, we just need to eval it (in case of '|' usage)
        if attrs.is_of_type(str):
            return yaml.load(self.jinja_env.from_string(attrs.value).render(ctx), Loader=UniqueKeyLoader)

        # if the section is not a string, i.e. values can contain jinja expressions,
        # we need to keep values to eval them later
        return attrs

    def eval_rec(self, name, val_node, ctx, file, line_number):
        """
        Evaluate structure value recursively using given context
        """
        if has_type(val_node, dict):
            return {k: self.eval_rec(name, v, ctx, file, line_number) for k, v in val_node.items()}
        if has_type(val_node, list):
            return [self.eval_rec(name, i, ctx, file, line_number) for i in val_node]
        if has_type(val_node, str):
            try:
                return self.jinja_env.from_string(to_value(val_node)).render(ctx)
            except TemplateSyntaxError as e:
                Error.critical(
                    f"Jinja syntax error in vars.{name} - {str(e)}",
                    file=file,
                    line=line_number)
            except Exception as e:
                Error.critical(f"Jinja render error in vars.{name} - {e}",
                               file,
                               line_number)
        return to_value(val_node)

    def eval_var_value(self, prop, val_node, ctx, att_name, location):
        """
        Method to eval/load variable value according to its type
        """
        file = line_number = None
        if location:
            # get file and line_number from location if it's passed
            file = location.file_name
            line_number = location.line_number
        elif isinstance(val_node, YamlNode):
            # otherwise try to take file and line_number from node if available
            file = val_node.file
            line_number = val_node.line_number

        val = to_value(val_node)
        # we get actual type from 'type' parameter if it is defined, otherwise it is str
        actual_type = get_var_real_type(to_value(prop.get('type'))) or str
        # if 'type' is not defined and default value is null, actual_type still can be None
        # we still check val since for some languages/platforms it can be null
        if actual_type and val is not None:
            # we evaluate jinja expression when type is str, or when we have type mismatch
            if actual_type is str or not has_type(val, actual_type):
                if not has_type(val, str):
                    Error.error(f"Type mismatch for '{att_name}' variable: "
                                f"it has value of {type(val)} type instead of required {actual_type}",
                                file,
                                line_number)
                val = self.eval_rec(att_name, val, ctx, file, line_number)

                # we load evaluated result to yaml if we have type mismatch but type is not str
                if actual_type is not str:
                    val = yaml.load(val, Loader=UniqueKeyLoader)
            else:
                # if we have type match, we need to call recursive evaluation function
                val = self.eval_rec(att_name, val, ctx, file, line_number)

        # we put single string element inside of list if 'type' attribute of variable is list
        if actual_type is list and isinstance(val, str):
            val = [val]
        # validate result type with actual type
        # we have cases when 'type' is specified but value can be null, that's why we need to check also val is not None
        elif val is not None and not isinstance(val, actual_type):
            Error.error(f"Type mismatch for '{att_name}' variable: "
                        f"it has value of {type(val)} type instead of required {actual_type}",
                        file,
                        line_number)

        return val



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.

"""
Module is responsible for context variables evaluating for current node.
"""
from collections import OrderedDict

from cppbind import default_config
from cppbind.common.error import Error
from cppbind.common.yaml_process import to_value
from cppbind.context_manager.var_eval import VariableEvaluator
from cppbind.ir.ast import (
    Node,
    RootNode,
    ROOT_KIND_NAME,
    DIR_KIND_NAME,
    FILE_KIND_NAME
)
from cppbind.parser.cppbind_api_parser import APIParser
from cppbind.utils import init_jinja_env

ALL_PLATFORMS = sorted(list(default_config.platforms))


class ContextManager:
    """
    A class for evaluating current context variables
    using current context to assign the result to the current node.
    """

    def __init__(self, ctx_desc, platform, language):
        self.ctx_desc = ctx_desc
        self.platform = platform
        self.language = language
        jinja_env = init_jinja_env(language)
        self.cppbind_api_parser = APIParser(ctx_desc, jinja_env, platform, language)
        self.var_evaluator = VariableEvaluator(jinja_env)

    def eval_root_attrs(self, ctx, var_values, location=None):
        """Eval context variables for root node"""
        args = None
        api = Node.API_NONE

        # update context with var values to use them during evaluation
        ctx.update(var_values)

        parsed_api = self.cppbind_api_parser.parse_yaml_api(RootNode.ROOT_KEY, ctx)
        if parsed_api:
            api, args = parsed_api

        args = args or {}
        # overwrite parsed variables with command line values (it has higher priority)
        args.update(var_values)

        return api, self.__process_attrs(ROOT_KIND_NAME, args, location, ctx)

    def eval_dir_attrs(self, name, ctx, location=None):
        """Eval context variables for dir node"""
        api = args = None
        parsed_api = self.cppbind_api_parser.parse_yaml_api(name, ctx)
        if parsed_api:
            api, args = parsed_api
        return api, self.__process_attrs(DIR_KIND_NAME, args, location, ctx)

    def eval_file_attrs(self, name, ctx, location=None):
        """Eval context variables for file node"""
        parsed_api = self.cppbind_api_parser.parse_yaml_api(name, ctx)
        if parsed_api:
            api, args = parsed_api
            if api:
                return api, self.__process_attrs(FILE_KIND_NAME, args, location, ctx)
        return None

    def eval_clang_attrs(self, name, kind, api_section, ctx, location):
        """Eval context variables for clang node"""
        parsed_api = self.cppbind_api_parser.parse_api(name, api_section, location, ctx)
        if parsed_api:
            api, args = parsed_api
            if api:
                return api, self.__process_attrs(kind, args, location, ctx)
        return None

    def __process_attrs(self, kind, args, location, ctx=None):
        """
        Evaluate/calculate current context variables values by its' kind and current context.
        """
        ctx = ctx or OrderedDict()
        args = args or OrderedDict()
        res = OrderedDict()

        # dummy object for identifying undefined and null value defined variables 
        undefined = object()

        # add all missing attributes
        for att_name, properties in self.ctx_desc.get_var_def().items():
            new_att_val = args.get(att_name, undefined)

            allowed = kind in properties["allowed_on"]
            if new_att_val is undefined:
                # check mandatory attribute existence
                if kind in properties["required_on"]:
                    Error.error(f"Attribute '{att_name}' is mandatory attribute on {kind}.",
                                location.file_name if location else None,
                                location.line_number if location else None)
                    break

                # inherit from parent or add default value
                if properties.get('inheritable'):
                    # directory based nodes may not have parent
                    if ctx:
                        new_att_val = ctx.get(att_name, undefined)

                if allowed:
                    if new_att_val is undefined:
                        # use default value
                        new_att_val = ContextManager.get_attr_default_value(
                            properties, self.platform, self.language, undefined)

                        if new_att_val is not undefined:
                            new_att_val = self.var_evaluator.eval_var_value(properties,
                                                                            new_att_val,
                                                                            ctx,
                                                                            att_name,
                                                                            location)
            else:
                # attribute is set check weather or not it is allowed.
                if not allowed:
                    Error.error(f"Attribute {att_name} is not allowed on {kind}.",
                                location.file_name if location else None,
                                location.line_number if location else None)
                    break

                new_att_val = self.var_evaluator.eval_var_value(properties,
                                                                new_att_val,
                                                                ctx,
                                                                att_name,
                                                                location)

            # validate variable values
            if new_att_val not in (None, undefined):
                self.cppbind_api_parser.validate_attr(att_name, new_att_val)

            options = to_value(ContextManager.get_var_property_value(properties, self.platform, self.language, 'options'))

            if new_att_val not in (None, undefined) and options is not None:
                if new_att_val not in options:
                    Error.error(f"Value mismatch for '{att_name}' variable: "
                                f"value {new_att_val} is not in list of allowed options",
                                location.file_name if location else None,
                                location.line_number if location else None)

            if new_att_val is not undefined:
                # add attr to current node context so that it can be used for coming attributes
                ctx[att_name] = new_att_val
                res[att_name] = new_att_val

        return res

    @staticmethod
    def get_attr_default_value(props, plat, lang, default=None):
        return ContextManager.get_var_property_value(props, plat, lang, 'default', default)

    @staticmethod
    def get_var_property_value(props, plat, lang, prop, default=None):
        """
        Retrieve language/platform specific default value for current variable.
        """

        # detect illegal specification of plat/lang option
        if plat + f'.{prop}' in props and lang + f'.{prop}' in props:
            Error.critical(
                f"Conflict of attributes in attributes definition file: {plat}.default and {lang}.default: "
                f"only one of them must be defined separately, or they must be both specified")

        # if plat/lang specific key is preset, return corresponding section
        # we search for specific key by descending order of priority
        for key in (plat + '.' + lang + f'.{prop}', plat + f'.{prop}', lang + f'.{prop}', prop):
            if key in props:
                return props[key]

        return default

    def filter_by_plat_lang(self, var_values):
        """
        Filter current platform/language specific values from initial context provided via command line arguments
        """
        res = {}

        if var_values is None:
            return res

        # convert argparse.Namespace into dict be able to lookup and iterate over it
        var_values = var_values.__dict__

        # pick those variable names which are present in variable definitions
        var_def = self.ctx_desc.get_var_def()
        var_names = set(name.split('.')[-1] for name in var_values if name in var_def)

        for name in var_names:
            prop = var_def[name]
            if 'cmd_line' not in prop['allowed_on']:
                continue

            plat_lang_opt = f"{self.platform}.{self.language}.{name}"
            plat_opt = f"{self.platform}.{name}"
            lang_opt = f"{self.language}.{name}"

            # search for value from highest to lowest priority (plat+lang+name, plat+name, lang+name, name)
            for opt in (plat_lang_opt, plat_opt, lang_opt, name):
                if opt in var_values and var_values[opt] is not None:
                    res[name] = var_values[opt]
                    # no need for later searching since we already found an option with the highest possible priority
                    break

        return res



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.


"""Module to define cxx related element class and its dependencies"""

import types

from cached_property import cached_property

import clang.cindex as cli
import cppbind.utils.clang as cutil
from cppbind.utils import DefaultValueKind
from .cxx_type import CXXType


class CXXElement:
    """Wrapper base type on clang cursor to provide unified access to cxx functionality/properties"""

    def __init__(self, clang_cursor):
        self._clang_cursor = clang_cursor

    @property
    def type(self):
        """Get the type of the current cursor"""
        return CXXType(self._clang_cursor.type)

    def get_full_displayname(self):
        """The full displayname of the cursor"""
        return cutil.get_full_displayname(self._clang_cursor)

    def get_signature(self):
        """The signature of the cursor"""
        return cutil.get_signature(self._clang_cursor)

    def get_template_parameter_cursors(self):
        """Returns the list of template parameter cursors for the clang cursor"""
        return cutil.get_template_parameter_cursors(self._clang_cursor)

    @property
    def kind_name(self):
        """Kind name of the cursor"""

        cl_kind = self._clang_cursor.kind.name.lower().replace("_decl", "").replace("cxx_", "")
        if self._clang_cursor.kind == cli.CursorKind.FUNCTION_TEMPLATE:
            if self._clang_cursor.lexical_parent.kind in (cli.CursorKind.STRUCT_DECL,
                                                          cli.CursorKind.CLASS_DECL,
                                                          cli.CursorKind.CLASS_TEMPLATE,
                                                          cli.CursorKind.CLASS_TEMPLATE_PARTIAL_SPECIALIZATION):
                if self._clang_cursor.spelling == self._clang_cursor.lexical_parent.spelling:
                    return 'constructor_template'
                else:
                    return 'method_template'

        return cl_kind

    @property
    def name(self):
        """Spelling of the cursor"""
        return self._clang_cursor.spelling

    @property
    def displayname(self):
        """Display name of the cursor"""
        return self._clang_cursor.displayname

    @property
    def source_file_name(self):
        """The source file path of the cursor"""
        return self._clang_cursor.extent.start.file.name

    @property
    def is_public(self):
        """Whether the cursor accessibility is public"""
        return self._clang_cursor.access_specifier == cli.AccessSpecifier.PUBLIC

    @property
    def is_private(self):
        """Whether the cursor accessibility is private"""
        return self._clang_cursor.access_specifier == cli.AccessSpecifier.PRIVATE

    @property
    def is_protected(self):
        """Whether the cursor accessibility is protected"""
        return self._clang_cursor.access_specifier == cli.AccessSpecifier.PROTECTED

    @property
    def raw_comment(self):
        """Returns the raw comment text associated with the cursor"""
        return self._clang_cursor.raw_comment

    @cached_property
    def namespace(self):
        """The namespace of the cursor"""

        namespaces = []
        parent = self._clang_cursor.lexical_parent
        while parent and parent.kind == cli.CursorKind.NAMESPACE:
            if parent.spelling:
                namespaces.append(parent.spelling)
            parent = parent.lexical_parent

        return '::'.join(reversed(namespaces))

    @property
    def is_abstract(self):
        """Checks whether the cursor is an abstract record"""
        return self._clang_cursor.is_abstract_record()

    @property
    def is_open(self):
        """Checks whether the function cursor is open"""
        return not cutil.is_final_cursor(self._clang_cursor)

    @property
    def is_templated(self):
        """Checks whether the cursor is templated"""
        return cutil.is_templated(self._clang_cursor)

    @cached_property
    def template_parameters(self):
        """Return the list of namespaces containing template parameter information.
         For example for
          template <typename T, typename... Ts> class tuple {}
        will return:
        [
            SimpleNamespace(is_variadic=False, kind=CursorKind.TEMPLATE_TYPE_PARAMETER, spelling='T'),
            SimpleNamespace(is_variadic=True, kind=CursorKind.TEMPLATE_TYPE_PARAMETER, spelling='Ts'),
        ]"""

        return [types.SimpleNamespace(
            is_variadic=cutil.is_variadic(c),
            kind=c.kind,
            spelling=c.spelling
        ) for c in self.get_template_parameter_cursors()]

    @property
    def parent(self):
        """
        Returns:
            Lexical parent of current element.
        """

        lexical_parent = self._clang_cursor.lexical_parent
        return CXXElement.create_cxx_element(lexical_parent) if lexical_parent else None

    @staticmethod
    def create_cxx_element(clang_cursor):
        """Static factory method to create a cxx element based on the cursor kind"""

        kind = clang_cursor.kind
        if kind == cli.CursorKind.ENUM_DECL:
            return CXXEnumElement(clang_cursor)
        if kind in (
                cli.CursorKind.FUNCTION_DECL,
                cli.CursorKind.FUNCTION_TEMPLATE,
                cli.CursorKind.CXX_METHOD,
                cli.CursorKind.CONSTRUCTOR
        ):
            return CXXFunctionElement(clang_cursor)
        if kind in (
                cli.CursorKind.CLASS_DECL,
                cli.CursorKind.STRUCT_DECL,
                cli.CursorKind.CLASS_TEMPLATE,
                cli.CursorKind.CLASS_TEMPLATE_PARTIAL_SPECIALIZATION
        ):
            return CXXClassElement(clang_cursor)
        if kind == cli.CursorKind.PARM_DECL:
            return CXXArgumentElement(clang_cursor)
        if kind == cli.CursorKind.FIELD_DECL:
            return CXXMemberElement(clang_cursor)

        return CXXElement(clang_cursor)

    def __getattr__(self, prop):
        """
        Delegate all missing properties to the underlying clang Cursor object.
        This is done to eliminate the usages of cursor and use cxx element instead of it.
        """

        if not hasattr(self._clang_cursor, prop):
            raise AttributeError(f"{self.__class__.__name__}.{prop} is invalid")

        return getattr(self._clang_cursor, prop)


class CXXEnumElement(CXXElement):
    """Wrapper type on clang enum cursor"""

    def __init__(self, clang_cursor):
        super(CXXEnumElement, self).__init__(clang_cursor)

    @property
    def type_name(self):
        return self._clang_cursor.type.spelling

    @cached_property
    def enum_cases(self):
        """Information about enum cases"""

        _values = []
        last_case_comment = None
        for enum_value_c in self._clang_cursor.walk_preorder():
            if enum_value_c.kind != cli.CursorKind.ENUM_CONSTANT_DECL:
                continue
            if enum_value_c.raw_comment != last_case_comment:
                comment = cutil.extract_pure_comment(enum_value_c.raw_comment)
            elif last_case_comment:
                comment = ['', 'The same as previous case comment.', '']
            else:
                comment = None
            last_case_comment = enum_value_c.raw_comment
            enum_val_params = types.SimpleNamespace(name=enum_value_c.spelling,
                                                    value=enum_value_c.enum_value,
                                                    comment=comment)
            _values.append(enum_val_params)

        return _values


class CXXClassElement(CXXElement):
    """Wrapper type on clang class/struct cursor"""

    def __init__(self, clang_cursor):
        super(CXXClassElement, self).__init__(clang_cursor)

    @property
    def type_name(self):
        return self._clang_cursor.type.spelling

    @property
    def is_polymorphic(self):
        """Checks the cursor corresponds to a polymorphic type"""
        return cutil.is_polymorphic(self._clang_cursor)

    @property
    def has_multiple_base_branches(self):
        """Checks the class/struct has multiple base branches in its hierarchy"""
        return cutil.has_multiple_base_branches(self._clang_cursor)

    @property
    def is_template(self):
        """Checks whether the class/struct cursor corresponds to a template"""
        return self._clang_cursor.kind in (
            cli.CursorKind.CLASS_TEMPLATE, cli.CursorKind.CLASS_TEMPLATE_PARTIAL_SPECIALIZATION)

    @property
    def base_type_elements(self):
        """
        Returns:
            List of base type elements.
        """

        _base_type_elements = []
        cursor = self._clang_cursor
        if self._clang_cursor.kind == cli.CursorKind.CXX_BASE_SPECIFIER:
            cursor = self._clang_cursor.get_definition()
        for base_specifier in cursor.get_children():
            if base_specifier.kind == cli.CursorKind.CXX_BASE_SPECIFIER:
                _base_type_elements.append(CXXClassElement(base_specifier))

        return _base_type_elements

    @property
    def ancestors(self):
        """
        Returns:
            List of ancestor elements.
        Raises:
            AttributeError: If current element is not a class/struct.
        """

        def walk(base_type_elements):
            for base in base_type_elements:
                for _base in walk(base.base_type_elements):
                    yield _base
                yield base

        return list(walk(self.base_type_elements))

    @cached_property
    def base_types(self):
        """
        Returns:
            List of base types.
        """
        return [element.type for element in self.base_type_elements]


class CXXFunctionElement(CXXElement):
    """Wrapper type on clang function/method cursor"""

    def __init__(self, clang_cursor):
        super(CXXFunctionElement, self).__init__(clang_cursor)

    @property
    def is_template(self):
        """Checks whether the function cursor corresponds to a template"""
        return self._clang_cursor.kind == cli.CursorKind.FUNCTION_TEMPLATE

    @property
    def is_static(self):
        """Checks whether the function cursor is static"""
        return self._clang_cursor.is_static_method()

    @property
    def is_const(self):
        """Checks whether the function cursor is const"""
        return self._clang_cursor.is_const_method()

    @property
    def is_overloaded(self):
        """Checks the function cursor is overloaded"""
        return cutil.is_overloaded(self._clang_cursor)

    @cached_property
    def overridden_elements(self):
        """
        Returns:
            List of overridden elements.
        """

        def _get_overridden_elems(cxx_element):

            elements = []
            if cxx_element.kind == cli.CursorKind.CXX_METHOD and cxx_element.get_overriden_cursors():
                for overridden in cxx_element.get_overriden_cursors():
                    elements.append(CXXFunctionElement(overridden))
                    elements += _get_overridden_elems(overridden)
            return elements

        return _get_overridden_elems(self)

    @property
    def result_type(self):
        """Return type info for function"""

        if self._clang_cursor.kind == cli.CursorKind.CONSTRUCTOR:
            return None

        return CXXType(self._clang_cursor.result_type)

    @property
    def args(self):
        """Argument elements of the function element"""

        _args = []

        # for function templates Cursor.get_arguments returns an empty array,
        # using get_children instead
        if self._clang_cursor.kind == cli.CursorKind.FUNCTION_TEMPLATE:
            for i, arg_c in enumerate(self._clang_cursor.get_children()):
                if arg_c.kind == cli.CursorKind.PARM_DECL:
                    # pass arg_idx to CXXArgumentElement to use arg{i} name instead of spelling when spelling is empty
                    _args.append(CXXArgumentElement(arg_c, arg_idx=i))
        else:
            for i, arg_c in enumerate(self._clang_cursor.get_arguments()):
                # pass arg_idx to CXXArgumentElement to use arg{i} name instead of spelling when spelling is empty
                _args.append(CXXArgumentElement(arg_c, arg_idx=i))

        return _args


class CXXArgumentElement(CXXElement):
    """Wrapper type on clang argument/parameter cursor"""

    def __init__(self, clang_cursor, arg_idx=None):
        super(CXXArgumentElement, self).__init__(clang_cursor)
        self.__arg_idx = arg_idx

    @property
    def name(self):
        """The name of the argument"""

        spelling = self._clang_cursor.spelling

        # use arg{i} name when the cursor spelling is empty
        # TODO: maybe we can move this logic to snippets
        if (self.is_variadic or not spelling) and self.__arg_idx is not None:
            return f'{spelling or "arg"}{self.__arg_idx + 1}'

        return spelling

    def __get_default(self, param_var):
        """Utility method to get default parameter kind and value"""

        # todo implementation is odd for now
        val = None
        kind = None
        if '=' in [p.spelling for p in param_var.get_tokens()]:
            for def_curs in param_var.walk_preorder():
                if def_curs.kind in (
                        cli.CursorKind.INTEGER_LITERAL,
                        cli.CursorKind.FLOATING_LITERAL,
                        cli.CursorKind.IMAGINARY_LITERAL,
                        cli.CursorKind.STRING_LITERAL,
                        cli.CursorKind.CHARACTER_LITERAL,
                        cli.CursorKind.CXX_BOOL_LITERAL_EXPR
                ):
                    kind = DefaultValueKind.LITERAL
                    val = next(def_curs.get_tokens(), None)
                    if val:
                        val = val.spelling
                elif def_curs.kind in (
                        cli.CursorKind.CXX_NULL_PTR_LITERAL_EXPR,
                        cli.CursorKind.GNU_NULL_EXPR,
                        cli.CursorKind.NULL_STMT):
                    val = 'nullptr'
                    kind = DefaultValueKind.NULL_PTR
                elif def_curs.kind == cli.CursorKind.DECL_REF_EXPR:
                    val = ''.join([token.spelling for token in def_curs.get_tokens()])
                    kind = DefaultValueKind.ENUM
                elif def_curs.kind == cli.CursorKind.CALL_EXPR:
                    kind = DefaultValueKind.CALL_EXPR
                    tokens = list(def_curs.get_tokens())
                    val = ''.join([token.spelling for token in tokens if token.spelling != '='])

                    if len(tokens) == 1:
                        if tokens[0].kind == cli.TokenKind.LITERAL:
                            # std::string case
                            kind = DefaultValueKind.LITERAL
                        elif val == 'nullptr':
                            # std::shared_ptr case
                            kind = DefaultValueKind.NULL_PTR
                    # value is retrieved break to not override it
                    break
        return kind, val

    @cached_property
    def __default_info(self):
        """Calculates default value information"""
        default_kind, default_value = self.__get_default(self._clang_cursor)
        return types.SimpleNamespace(kind=default_kind, value=default_value)

    @cached_property
    def default(self):
        """Default value of the default argument"""
        return self.__default_info.value

    @cached_property
    def default_is_literal(self):
        """Checks whether a default value is a literal"""
        return self.__default_info.kind == DefaultValueKind.LITERAL

    @cached_property
    def default_is_nullptr(self):
        """Checks whether a default value is nullptr"""
        return self.__default_info.kind == DefaultValueKind.NULL_PTR

    @cached_property
    def is_variadic(self):
        """Return whether the argument is template parameter pack or not."""
        return cutil.is_variadic(self._clang_cursor)


class CXXMemberElement(CXXElement):
    """Wrapper type on clang member/field cursor"""

    def __init__(self, clang_cursor, arg_idx=None):
        super(CXXMemberElement, self).__init__(clang_cursor)



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.

"""
Helper codes for python conversion
"""
import os

from isort.api import sort_code_string

from . import *

OPERATOR_MAPPING = {
    '*': '__mul__',
    '*=': '__imul__',
    '-': '__sub__',
    '-=': '__isub__',
    '+': '__add__',
    '+=': '__iadd__',
    '/': '__truediv__',
    '/=': '__itruediv__',
    '<': '__lt__',
    '<=': '__le__',
    '>': '__gt__',
    '>=': '__ge__',
    '==': '__eq__',
    '!=': '__ne__',
    '%': '__mod__',
    '&': '__and__',
    '|': '__or__',
    '|=': '__ior__',
    '^': '__xor__',
    '^=': '__ixor__',
    '~': '__invert__',
    '~=': '__iinvert__',
    '<<': '__lshift__',
    '>>': '__rshift__',
    '<<=': '__ilshift__',
    '>>=': '__irshift__',
    '&=': '__iand__',
    '%=': '__imod__',
    'bool': '__bool__',
    '()': '__call__',
    '[]': '__getitem__',
}

sort_imports = sort_code_string


def get_operator_name(spelling):
    operator = spelling.replace('operator', '').strip()
    return OPERATOR_MAPPING.get(operator, spelling)



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.

"""
Common utils that can by used from different modules
"""
import enum
import errno
import glob
import importlib.util
import logging
import os
import re
import shutil
import sys
from collections.abc import Iterable
from functools import cmp_to_key, lru_cache
from types import SimpleNamespace

from isort.api import sort_code_string
from jinja2 import BaseLoader, Environment, StrictUndefined
from wcmatch import glob as wcglob

from cppbind import BANNER_LOGO
from cppbind import default_config
from cppbind.common import JINJA_UNIQUE_MARKER, YAML_CONFIG_TEMPLATE_PATH
from cppbind.common.error import Error
from cppbind.ir import ElementKind

DEFAULT_HELPER = 'cppbind_helper'


class DefaultValueKind(enum.IntEnum):
    CALL_EXPR = 1
    ENUM = 2
    LITERAL = 3
    NULL_PTR = 4


def absolute_path_from_glob(src_glob):
    """
    Finds all the absolute file paths matching a specified pattern
    Args:
        src_glob(list): List of glob paths
    Returns:
        list: List of absolute paths
    """
    all_absolute_paths = set()
    for file in src_glob:
        abs_paths = (os.path.abspath(file_path)
                     for file_path in glob.glob(file.strip(), recursive=True))
        all_absolute_paths.update(abs_paths)

    return all_absolute_paths


def list_files(path, extentions=None):
    """ List all files in a directory specified by path
    Args:
        path - the root directory path
        extensions - a iterator of file extensions to include, pass None to get all files.
    Returns:
        A list of files specified by extensions
    """
    file_paths = []
    for root, _, files in os.walk(path):
        for file in files:
            if extentions is None:
                file_paths.append(os.path.join(os.path.abspath(root), file))
            else:
                for ext in extentions:
                    if file.endswith(ext):
                        file_paths.append(os.path.join(os.path.abspath(root), file))

    return file_paths


def filter_by_glob(files, pattern):
    """ Filter list of files by specified pattern
    Args:
        files - list of file paths
        pattern - filter that describes a set of strings that matches the pattern
    Returns:
        A list of files filtered by specified pattern
    """
    return [file for file in files if wcglob.globmatch(file, pattern, flags=wcglob.GLOBSTAR)]


def load_from_paths(loader, path_name, default_dirs):
    for dir_ in default_dirs:
        try:
            return loader(os.path.join(dir_, path_name))
        except FileNotFoundError:
            continue
    raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), path_name)


def load_module_from_path(module_name, path_name):
    module_spec = importlib.util.spec_from_file_location(module_name, path_name)
    module = importlib.util.module_from_spec(module_spec)
    module_spec.loader.exec_module(module)
    return module


def load_module_from_paths(module_name, path_name, default_dirs):
    return load_from_paths(lambda path: load_module_from_path(module_name, path),
                           path_name, default_dirs)


def load_rule_module(language, rule, default_dirs):
    module = load_module_from_paths(f"{language}.rule",
                                    rule,
                                    default_dirs)
    module.set_language(language)
    return module


def get_host_platform():
    platform = sys.platform
    if platform.startswith('linux'):
        return 'linux'
    if platform.startswith('win'):
        return 'win'
    if platform == 'darwin':
        return 'mac'

    raise Exception(f"Not supported os platform: {platform}")


def get_android_ndk_sysroot(ndk_path):
    """
    Construct android ndk sysroot path for target platform
    """
    platform = get_host_platform()

    # now we check only for mac, win, linux platforms, but we also have ios, android platforms (also there is win32)
    platform_section = 'linux-x86_64'
    if platform == 'mac':
        platform_section = 'darwin-x86_64'
    elif platform == 'win':
        platform_section = 'windows-x86_64'

    sysroot_path = os.path.join(ndk_path if ndk_path is not None else '/usr/local/share/android-ndk',
                                'toolchains/llvm/prebuilt', platform_section, 'sysroot')

    if not os.path.isdir(sysroot_path):
        Error.critical(f"'ANDROID_NDK' environment variable is not set properly: "
                       f"{sysroot_path} directory doesn't exist")

    return sysroot_path


def make_snake_case(string, sub_strings=None):
    """
    Returns snake cased version of input string if sub_strings is None.
    If sub_strings is specified then only replaces all sub_string
    occurrences with their snake cased version.
    Args:
        string(str): String to replace
        sub_strings(list): List of sub_strings to be replaced with their snake cased version.
    Returns:
        str: Snake cased input string.
    """
    if sub_strings:
        for p in sub_strings:
            string = string.replace(p, make_snake_case(p))
        return string
    return re.sub(r'(?<!^)(?=[A-Z])', '_', string).lower()


def make_camel_case(string, sub_strings=None, capitalize=False):
    """
    Returns camel cased version of input string if sub_strings is None.
    If sub_strings is specified then only replaces all sub_string
    occurrences with their camel cased version.
    Args:
        string(str): String to replace
        sub_strings(list): List of sub_strings to be replaced with their camel cased version.
    Returns:
        str: Camel cased input string.
    """
    if sub_strings:
        for p in sub_strings:
            string = string.replace(p, make_camel_case(p))
        return string
    init, *temp = string.strip('_').split('_')
    return ''.join([init.title() if capitalize else init, *map(str.title, temp)])


def clear_cppbind_generated_files(directory):
    """
    Traverses given directory and removes all files that contain CppBind banner.
    Args:
        directory(str): Directory path.
    """
    # remove all spaces from banner logo
    banner = re.sub(r'\s+', '', BANNER_LOGO)
    # We are using topdown=False to make sure that walk visits deepest
    # directories at first so no empty directory will be left.
    for root, _, filenames in os.walk(directory, topdown=False):
        for file in filenames:
            remove = False
            file_path = os.path.join(root, file)
            try:
                with open(file_path, 'r') as file:
                    content = file.read()
                    # remove all * and spaces from banner
                    content = re.sub(r'[*\s+]', '', content)
                    if banner in content:
                        remove = True
                if remove:
                    os.remove(file_path)
            except Exception as err:
                Error.warning(f"Cannot process {os.path.abspath(file_path)} file: {err}")
        if not os.listdir(root):
            try:
                os.rmdir(os.path.abspath(root))
            except OSError as err:
                Error.warning(f"Cannot delete {os.path.abspath(root)} directory: {err}")


def copy_yaml_config_template():
    """
    Copies cppbind yaml config template file containing default values into current directory.
    """
    cwd = os.getcwd()
    config_file_name = os.path.basename(YAML_CONFIG_TEMPLATE_PATH)
    if not os.path.exists(os.path.join(cwd, config_file_name)):
        shutil.copy(YAML_CONFIG_TEMPLATE_PATH, cwd)
    else:
        Error.critical(f"'{config_file_name}' project config file already exists in '{cwd}' directory. "
                       f"CppBind will skip config file initialisation.")


def get_var_real_type(type_name):
    """
    Returns real type of variable using 'type' parameter of variable property
    """
    if type_name is None:
        return None

    try:
        res = eval(type_name)
    except (NameError, SyntaxError, TypeError) as err:
        Error.critical(f"'{type_name.value}' cannot be used as a value for 'type' "
                       f"parameter since it's not python primitive type: {err}",
                       type_name.file,
                       type_name.line_number)

    if not isinstance(res, type):
        Error.critical(f"'{res}' cannot be used as a value for 'type' parameter since it's not python primitive type",
                       type_name.file,
                       type_name.line_number)

    return res


def get_public_attributes(attrs):
    """Filter and return only "public" attributes"""
    return {k: v for k, v in attrs.items() if not k.startswith('_')}


@lru_cache
def init_jinja_env(language):
    """
    Function which initializes jinja environment with custom filters/tests
    """

    def path_join(inputs_):
        return os.path.join(*inputs_)

    def format_list(inputs_, format_string, arg_name=None):
        if arg_name is not None:
            return [format_string.format(**{arg_name: data}) for data in inputs_]
        return [format_string.format(data) for data in inputs_]

    def join_unique(inputs_):
        return JINJA_UNIQUE_MARKER.join(inputs_)

    def match_regex(input_, *patterns):
        return any(re.match(pattern, input_) for pattern in patterns)

    def contains(input_, item):
        return item in input_

    def startswith(input_, prefix):
        return input_.startswith(prefix)

    def replace_regex(input_, pattern, repl, count=0):
        return re.sub(pattern, repl, input_, count)

    def _zip(*args):
        if len(args) < 2:
            Error.critical("'zip' filter requires at least 2 arguments")
        for arg in args:
            if not isinstance(arg, Iterable):
                Error.critical("Arguments of the 'zip' filter must be iterable objects")

        return (SimpleNamespace(**{f"_{idx + 1}": item for idx, item in enumerate(group)}) for group in zip(*args))

    def _split(inputs_):
        parts = []
        if not isinstance(inputs_, list):
            inputs_ = [inputs_]
        for input_ in inputs_:
            if isinstance(input_, str):
                parts += input_.split(JINJA_UNIQUE_MARKER)
            else:
                # input_ is a scope
                for part in input_.parts:
                    for part_ in str(part).split(JINJA_UNIQUE_MARKER):
                        if part_:
                            parts.append(part_)
        return parts

    def _default_comparator(a, b):
        return 1 if a > b else -1

    def sort_snippets(inputs_, cmp=_default_comparator, reverse=False):
        parts = _split(inputs_)
        return sorted(parts, key=cmp_to_key(cmp), reverse=reverse)

    def unique_snippets(inputs_):
        parts = _split(inputs_)
        return set(parts)

    def sort_python_code(source):
        return sort_code_string(source).strip()

    class DoxygenCommentStyle(enum.IntEnum):
        JAVADOC = 1
        CPP = 2

    def make_doxygen_comment(comment, style=DoxygenCommentStyle.JAVADOC):
        if not isinstance(style, DoxygenCommentStyle):
            raise ValueError(f'Incorrect doxygen style. Supported styles are: '
                             f'{",".join([f"DoxygenCommentStyle.{c.name}" for c in DoxygenCommentStyle])}')

        if isinstance(comment, str):
            comment = [comment]
        lines = []
        for line in comment:
            lines += line.split('\n')
        nl = '\n/// ' if style != DoxygenCommentStyle.JAVADOC else '\n * '
        if not lines or all((not line or line.isspace() for line in lines)):
            return ''
        if not lines[0] or lines[0].isspace():
            start = ''
        else:
            start = '/// ' if style != DoxygenCommentStyle.JAVADOC else nl
        _comment = f"""{start}{nl.join(lines)}"""
        return f"""{_comment}""" if style != DoxygenCommentStyle.JAVADOC else f"""/**{_comment}\n */"""

    def make_py_docstring(comment):
        nl = '\n'
        if isinstance(comment, str):
            comment = [comment]
        lines = []
        for line in comment:
            lines += line.split(nl)
        if not lines or all((not line or line.isspace() for line in lines)):
            return ""
        start = '' if not lines[0] or lines[0].isspace() else nl
        return f'"""{start}{nl.join(lines)}{nl}"""'

    def make_py_comment(comment):
        return '# ' + '\n# '.join(comment)

    def decapitalize(input_):
        return input_[0].lower() + input_[1:]

    def increment(input_, amount=1):
        return input_ + amount

    def map_callback(input_, callback, **kwargs):
        if isinstance(input_, Iterable):
            return [callback(item, **kwargs) for item in input_]
        return callback(input_, **kwargs)

    def map_type(input_):
        if isinstance(input_, Iterable):
            return [map_type(item) for item in input_]

        if not hasattr(input_, 'type'):
            Error.critical(f"'type' filter is acceptable only for objects (list of objects) with 'type' property. "
                           f"{type(input_)} doesn't have 'type' property.")

        return input_.type

    env = Environment(loader=BaseLoader(),
                      undefined=StrictUndefined,
                      extensions=['jinja2.ext.do', 'jinja2.ext.debug', 'jinja2.ext.loopcontrols'])

    env.filters['all'] = all
    env.filters['any'] = any
    env.filters['path_join'] = path_join
    env.filters['format_list'] = format_list
    env.filters['to_snake_case'] = make_snake_case
    env.filters['to_camel_case'] = make_camel_case
    env.filters['join_unique'] = join_unique
    env.filters['replace_regex'] = replace_regex
    env.filters['zip'] = _zip
    env.filters['sort_snippets'] = sort_snippets
    env.filters['unique_snippets'] = unique_snippets
    env.filters['sort_python_code'] = sort_python_code
    env.filters['make_doxygen_comment'] = make_doxygen_comment
    env.filters['make_py_docstring'] = make_py_docstring
    env.filters['make_py_comment'] = make_py_comment
    env.filters['decapitalize'] = decapitalize
    env.filters['increment'] = increment
    env.filters['map_callback'] = map_callback
    env.filters['type'] = map_type
    env.filters['set'] = lambda x: set(x)

    env.tests['match_regex'] = match_regex
    env.tests['contains'] = contains
    env.tests['startswith'] = startswith

    env.globals['Error'] = Error
    env.globals['DoxygenCommentStyle'] = DoxygenCommentStyle
    env.globals['path'] = os.path
    env.globals['pat_sep'] = os.sep
    env.globals['marker'] = JINJA_UNIQUE_MARKER
    env.globals['ElementKind'] = ElementKind

    # add default and custom helpers to jinja env
    for name, module in get_helper_modules(language).items():
        env.globals[name] = module

    return env


def get_language_helper_module(language):
    try:
        language_helper_module = importlib.import_module(f'cppbind.converter.{language}')
        logging.info(f"Loaded helper module for '{language}' language")
    except ModuleNotFoundError:
        language_helper_module = importlib.import_module('cppbind.converter')
        logging.info(f"Helper module is not found for '{language}' language, loading default instead")

    return language_helper_module


def get_helper_modules(language):
    default_module = get_language_helper_module(language)
    modules = {DEFAULT_HELPER: default_module}
    sys.modules[DEFAULT_HELPER] = default_module
    if hasattr(default_config.application, 'custom_helpers_dir'):
        helpers_paths = glob.glob(os.path.join(default_config.application.custom_helpers_dir, f'**/{language}/**/*.py'),
                                  recursive=True)
        for module_path in helpers_paths:
            module_name = os.path.splitext(os.path.basename(module_path))[0]
            spec = importlib.util.spec_from_file_location(module_name, os.path.abspath(module_path))
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            modules[module_name] = module

    return modules



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.
from .cppbind_api_parser import (
    TEMPLATE_TYPE_PARAMETER_KEY,
    TEMPLATE_NON_TYPE_PARAMETER_KEY,
    TEMPLATE_NAME_KEY,
    TEMPLATE_INSTANCE_KEY,
    TEMPLATE_KEY
)
